#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include "drivers/sb.h"
#include "paula.h"
#include "replayer.h" // Replayer_Tick()
#include "mixer.h"
#include "../hardware/video.h" // VBLANK_HZ
#include "audio.h"

static bool audioDetected;
static int16_t *lpFilterTabPtr;

// globals
bool audio16Bit, stereoAudio, audioOpen, audioAmiga500;
void *mixBufferL, *mixBufferR;
int8_t stereoSeparation = DEF_STEREO_SEP; // 0..128
int16_t *lpFilterTab, lastFilterSampleL, lastFilterSampleR;
uint32_t samplesPerTickHi, samplesPerTickLo, sampleCounterHi, sampleCounterLo, audioSideFactorMul;
audio_t audio;

void Audio_UpdateStereoSeparation(void)
{
	if (!stereoAudio)
		stereoSeparation = 0;
	else
		audioSideFactorMul = ((uint8_t)stereoSeparation * 128) / 100; // 0..100 -> 0..128
	
	Mixer_SetStereoMapping();
}

void Audio_ClearLpFilterState(void)
{
	lastFilterSampleL = lastFilterSampleR = 0;
}

static bool calcLpFilterTable(double audioFrequency)
{
#define TWO_PI 6.283185307179586
	
// Amiga 500 rev 6A cutoff Hz (1-pole 6dB/oct RC low-pass)
#define LP_CUTOFF_HZ 4420.970641441537

	double a, b;
	uint32_t mul32;

	lpFilterTabPtr = (int16_t *)malloc((65536 * sizeof (int16_t)) + 2);
	if (lpFilterTabPtr == NULL)
		return false;
		
	// align pointer (just in case)
	lpFilterTab = (int16_t *)(((uint32_t)lpFilterTabPtr + 1) & ~1);

	a = 2.0 - cos((LP_CUTOFF_HZ * TWO_PI) / audioFrequency);
	b = a - sqrt((a * a) - 1.0);

	// calculate 128kB LUT in fixed-point asm (for speed)
	mul32 = (uint32_t)((b * (UINT32_MAX+1.0)) + 0.5);
__asm
{
	pushad
	mov ebx,mul32
	not ebx	// b = 1.0 - b
	mov edi,lpFilterTab
	mov ecx,65535
	mov esi,32768 // 16-bit PCM sign flip (XOR)
L0:	mov eax,ecx
	xor eax,esi
	mul ebx
	mov [edi+ecx*2],dx
	dec ecx
	jns L0
	popad
};

	return true;
}

void Audio_SetSamplesPerTick(int64_t samples)
{
	samplesPerTickHi = samples >> 32;
	samplesPerTickLo = (uint32_t)samples;
}

void Audio_ClearSampleCounter(void)
{
	sampleCounterHi = sampleCounterLo = 0;
}

bool Audio_Detect(bool noSound)
{
	if (noSound)
	{
		audioDetected = false;
		return false;
	}
	
	if (SB_Detect())
	{
		audioDetected = true;
		return true;
	}
	else
	{
		switch (SB_GetError())
		{
			case SB_ERROR_NO_ENV_VAR:
				printf("WARNING: No BLASTER environment variable is set!");
			break;
			
			case SB_ERROR_INVALID_ENV_VAR:
				printf("WARNING: The BLASTER environment variable has invalid values!\n");
				printf("\n");
				printf("Accepted ranges are:\n");
				printf(" A (port)                 : 220..300\n");
				printf(" I (IRQ)                  : 1..15\n");
				printf(" D (8-bit DMA)            : 1..3\n");
				printf(" H (16-bit DMA, optional) : 1..7\n");
				printf("\n");
				printf("Here's a typical BLASTER environment variable for an SB16:\n");
				printf("SET BLASTER=A220 I5 D1 H5");
			break;
			
			case SB_ERROR_NO_CARD_DETECTED:
				printf("WARNING: Couldn't detect a Sound Blaster sound card!\n");
				printf("         If you really have a Sound Blaster card installed, make sure its\n");
				printf("         drivers are properly installed.");
			break;
			
			case SB_ERROR_UNSUPPORTED_CARD:
				printf("WARNING: Sound Blaster 1.0 is not supported!");
			break;
			
			default: break; // won't happen, but keeps compiler happy
		}
		
		printf("\n\nPress any key to continue with no sound...\n");
		getch();

		audioDetected = false;
		return false;
	}
}

bool Audio_Init(void)
{
	if (!audioDetected)
	{
		audio.dOutputFreq = VBLANK_HZ * 512; // extra precision needed
		return true; // "no audio" mode
	}
	
	if (!SB_Init())
		return false;

	mixBufferL = (int32_t *)calloc(AUDIO_BUF_SIZE, sizeof (int32_t));
	mixBufferR = (int32_t *)calloc(AUDIO_BUF_SIZE, sizeof (int32_t));

	if (mixBufferL == NULL || mixBufferR == NULL)
	{
		printf("ERROR: Out of memory!\n");
		return false; // mem gets free'd later
	}
	
	if (!calcLpFilterTable(audio.dOutputFreq))
	{
		printf("ERROR: Out of memory!\n");
		return false; // mem gets free'd later
	}
	
	Audio_ClearLpFilterState();	
	stereoSeparation = audio.defStereoSep;
	Audio_UpdateStereoSeparation();	

	audioOpen = true;
	return true;
}

bool Audio_StartDMA(void)
{
	if (!audioDetected)
		return true; // "no audio" mode

	SB_StartDMA();
	audio.soundReady = true;
	return true;
}

void Audio_Close(void)
{
	if (audio.soundReady)
	{
		SB_Close();
		audio.soundReady = false;
	}

	if (mixBufferL != NULL)
	{
		free(mixBufferL);
		mixBufferL = NULL;
	}
	
	if (mixBufferR != NULL)
	{
		free(mixBufferR);
		mixBufferR = NULL;
	}
	
	if (lpFilterTabPtr != NULL)
	{
		free(lpFilterTabPtr);
		lpFilterTabPtr = NULL;
	}
}
