// Sound Blaster 2.0 / Pro / 16 (or compatibles) driver for PT4DOS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include "sb.h"
#include "../audio.h"
#include "../../asm.h"
#include "../../hardware/dma.h"
#include "../../hardware/irq.h"

#define dspTimeConstant		0x40
#define dspBlockSize		0x48
#define dspTurnSpeakerON	0xD1
#define dspTurnSpeakerOFF	0xD3
#define dspGetDSPVersion	0xE1
#define dspMixerAddr		0x04
#define dspMixerData		0x05
#define dspReset		0x06
#define dspReadData		0x0A
#define dspWriteBuf		0x0C
#define dspDataAvail		0x0E
#define dspDMA8Ack		0x0E
#define dspDMA16Ack		0x0F
#define dspReady		0xAA
#define mixVocVol		0x04
#define mixMicVol		0x0A
#define mixOutFilter		0x0E

static bool sbInitialized;
static int8_t sbLastError, sbIrq, sbLoDma, sbHiDma, sbDma;
static uint8_t timeConstant, sbType, oldMicVol, oldVoiceVol, oldFilterStereo;
static int16_t sbPort;
static uint16_t dspVer;

// asb.asm
extern uint16_t sbDmaAckPort;
extern void __interrupt sbLoIrqInterrupt(void);
extern void __interrupt sbHiIrqInterrupt(void);

static bool processBlasterVariable(void)
{
	char ch, *p = getenv("BLASTER");
	if (p == NULL)
		return false;

	sbPort = sbIrq = sbLoDma = sbHiDma = -1;
	while (true)
	{
		// skip spaces
		while (*p == ' ')
			p++;

		ch = *p++;
		if (ch == '\0') // end of string?
			break;

		if (ch >= 'a' && ch <= 'z') // convert to upper case
			ch -= ('a' - 'A');

		switch (ch)
		{
			case 'A': sbPort  = strtoul(p, NULL, 16); break;		
			case 'I': sbIrq   = strtoul(p, NULL, 10); break;
			case 'D': sbLoDma = strtoul(p, NULL, 10); break;
			case 'H': sbHiDma = strtoul(p, NULL, 10); break;
			default: break;
		}
	}

	return true;
}

static void dspOut(uint8_t val)
{
	while (inp(sbPort+dspWriteBuf) & 0x80);
	outp(sbPort+dspWriteBuf, val);
}

static void dspIn(uint8_t *val)
{
	while (!(inp(sbPort+dspDataAvail) & 0x80));
	*val = inp(sbPort+dspReadData);
}

static bool SB_Reset(void)
{
	bool detected = false;

	__asm
	{
		push eax
		push ecx
		push edx

		cli

		mov dx,sbPort
		add dx,dspReset
		mov al,1
		out dx,al
		// ------------------------
		// wait at least 3 microseconds
		// ------------------------
		mov ecx,16
	L0:	in al,dx
		loop L0
		// ------------------------
		xor al,al
		out dx,al
		// ------------------------	
		mov dx,sbPort
		add dx,dspDataAvail	
		mov ecx,4000 // try 4000 times
	L1:	in al,dx
		test al,al
		js L2
		loop L1
		jmp done
	L2:	// ------------------------
		sub dx,4
		in al,dx
		cmp al,0AAh
		jne done
		mov byte ptr [detected],1
	done:	// ------------------------	

		sti

		pop edx
		pop ecx
		pop eax
	}
	
	return detected;
}

bool SB_Detect(void)
{
	uint8_t val, hi, lo;
	bool detected = false;
	
	sbLastError = SB_ERROR_NONE;
	
	sbType = SBUnknown;

	if (!processBlasterVariable())
	{
		sbLastError = SB_ERROR_NO_ENV_VAR;
		return false;
	}
	
	// check if we have the required BLASTER parameters
	if (sbPort == -1 || sbIrq == -1 || (sbLoDma == -1 && sbHiDma == -1))
	{
		sbLastError = SB_ERROR_INVALID_ENV_VAR;
		return false; 
	}

	// check if the BLASTER parameters are valid
	if (sbPort < 0x220 || sbPort > 0x300 ||
		sbLoDma == 0 || (sbLoDma != -1 && sbLoDma >= 4) ||
		sbHiDma == 0 || (sbHiDma != -1 && sbHiDma > 7) ||
		sbIrq < 1 || sbIrq > 15)
	{
		sbLastError = SB_ERROR_INVALID_ENV_VAR;
		return false; 
	}

	if (SB_Reset())
	{
		_disable();
		dspOut(dspGetDSPVersion);
		dspIn(&hi);
		dspIn(&lo);
		_enable();

		dspVer = (hi << 8) | lo;
		if (dspVer >= 0x400)
			sbType = SB16;
		else if (dspVer >= 0x300)
			sbType = SBPro;
		else if (dspVer >= 0x201)
			sbType = SB2;
		else
			sbType = SB1;
	}
	else
	{
		sbLastError = SB_ERROR_NO_CARD_DETECTED;
		return false;
	}

	if (sbType == SB1) // we need >31388Hz for PT4DOS
	{
		sbLastError = SB_ERROR_UNSUPPORTED_CARD;
		return false;
	}

	return true;
}

static void SetMixer(void)
{
	_disable();

	// read old volumes
	outp(sbPort+dspMixerAddr, mixMicVol);
	oldMicVol = inp(sbPort+dspMixerData);
	outp(sbPort+dspMixerAddr, mixVocVol);
	oldVoiceVol = inp(sbPort+dspMixerData);

	// mute microphone input
	outp(sbPort+dspMixerAddr, mixMicVol);
	outp(sbPort+dspMixerData, 0x00);

	// set max voice volume
	outp(sbPort+dspMixerAddr, mixVocVol);
	outp(sbPort+dspMixerData, 0xFF);

	if (sbType == SBPro)
	{
		// read old stereo/filter mode
		outp(sbPort+dspMixerAddr, mixOutFilter);
		oldFilterStereo = inp(sbPort+dspMixerData);

		// set mono + no filter
		outp(sbPort+dspMixerAddr, mixOutFilter);
		outp(sbPort+dspMixerData, 0x20);
	}

	_enable();
}

static void RestoreMixer(void)
{
	_disable();

	// set back old microphone volume
	outp(sbPort+dspMixerAddr, mixMicVol);
	outp(sbPort+dspMixerData, oldMicVol);

	// set back old voice volume
	outp(sbPort+dspMixerAddr, mixVocVol);
	outp(sbPort+dspMixerData, oldVoiceVol);

	if (sbType == SBPro)
	{
		// set back old stereo/filter mode
		outp(sbPort+dspMixerAddr, mixOutFilter);
		outp(sbPort+dspMixerData, oldFilterStereo);
	}

	_enable();
}

bool SB_StartDMA(void)
{
	dspOut(dspTurnSpeakerON);

	if (!DMA_Start(sbDma, DMA_PAUSE_IRQS))
		return false;

	_disable();
	if (sbType == SB16)
	{
		dspOut(dspTimeConstant);
		dspOut(timeConstant);

		dspOut(0xB6); // 16-bit output
		dspOut(0x30); // 16-bit stereo signed PCM

		// block size (highest possible for less IRQ pollution)
		dspOut(0xFF);
		dspOut(0xFF);
	}
	else // SB 2.0 / SB Pro / SB Pro 2.0
	{
		dspOut(dspTimeConstant);
		dspOut(timeConstant);

		// block size (highest possible for less IRQ pollution)
		dspOut(dspBlockSize);
		dspOut(0xFF);
		dspOut(0xFF);

		dspOut(0x90); // 8-bit mono unsigned PCM (auto-init high-speed output)
	}
	_enable();

	return true;
}

bool SB_Init(void) // must be ran after SB_Detect()
{
	sbInitialized = false;
	stereoAudio = audio16Bit = false;
	sbDma = sbLoDma;

	if (sbType == SB2)
	{
		timeConstant = 232; // ~41666.67Hz @ mono (max on SB 2.0, unfortunately)
	}
	else if (sbType == SBPro)
	{
		timeConstant = 233; // ~43478.26Hz @ mono (234 possible, but not on (some?) clones)
	}
	else // SB16
	{
		timeConstant = 234; // ~45454.55Hz @ stereo
	
		audio16Bit = true;
		stereoAudio = true; // (note: we only accept stereo when at 16-bit output)

		if (sbHiDma != -1) // use high DMA if present
			sbDma = sbHiDma;
	}

	if (!DMA_Init(sbDma, AUDIO_BUF_SIZE, DMA_MODE_READ | DMA_MODE_AUTOINIT | DMA_MODE_SINGLE))
	{
		printf("ERROR: Couldn't initialize audio DMA!\n");
		return false;
	}

	audio.dmaBuffer = DMA_GetBuffer(sbDma);
	if (audio16Bit)
		ASM_FastFill32(audio.dmaBuffer, 0x00000000, AUDIO_BUF_SIZE/4); // signed 16-bit
	else
		ASM_FastFill32(audio.dmaBuffer, 0x80808080, AUDIO_BUF_SIZE/4); // unsigned 8-bit

	audio.dOutputFreq = 1000000.0 / (256 - timeConstant);
	audio.dmaNum = sbDma;

	if (sbType == SBPro || sbType == SB16)
		SetMixer();

	// for asb.asm
	sbDmaAckPort = audio16Bit ? (sbPort+dspDMA16Ack) : (sbPort+dspDMA8Ack); 

	// for apitint.asm
	audio.sampleShift = audio16Bit ? 1 : 0;
	if (stereoAudio)
		audio.sampleShift++;

	IRQ_SetIntVector(sbIrq, (sbIrq < 8) ? sbLoIrqInterrupt : sbHiIrqInterrupt);
	IRQ_Enable(sbIrq);

	sbInitialized = true;
	return true;
}

void SB_Close(void)
{
	if (!sbInitialized)
		return;

	if (sbType == SBPro || sbType == SB16)
		RestoreMixer();

	IRQ_RestoreIntVector(sbIrq);
	DMA_Close(sbDma);

	SB_Reset();

	if (sbPort != 0)
		dspOut(dspTurnSpeakerOFF);
}

uint8_t SB_GetModel(void)
{
	return sbType;
}

uint8_t SB_GetError(void)
{
	return sbLastError;
}
