#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h> // disable()/enable()
#include "replayer.h"
#include "audio.h"
#include "paula.h"
#include "../asm.h"
#include "../ui.h" // ui.RedrawPatternFlag
#include "../scopes.h" // SCOPE_W
#include "../sampler.h"
#include "../hardware/mouse.h" // Mouse_UpdateColor()

#define CIA_PAL_CLK 709379

// -----------------------------------------------------------------------------
// ----------------------------------- MACROS ----------------------------------
#define SWAP16(x) \
( \
	(((uint16_t)(x & 0x00FF)) << 8) | \
	(((uint16_t)(x & 0xFF00)) >> 8)   \
)

#define SEEK_TO(x) if (fseek(f, x, SEEK_SET) != 0) goto errIO
#define READ_BYTES(p, n) if (fread(p, 1, n, f) != n) goto errIO
#define READ_BYTE(p) if (fread(&p, 1, 1, f) != 1) goto errIO
#define READ_WORD(p) if (fread(&p, 2, 1, f) != 1) goto errIO
#define READ_LONG(p) if (fread(&p, 4, 1, f) != 1) goto errIO
#define READ_MOTOROLA_WORD(p) READ_WORD(p); p = SWAP16(p);
// -----------------------------------------------------------------------------

enum
{
	ID_FLT4  = 0x34544C46, // FLT4 (up to 64 patterns - non-PT)
	ID_MK64  = 0x2E4B2E4D, // M.K. (up to 64 patterns - PT1)
	ID_MK100 = 0x214B214D, // M!K! (65..100 patterns - PT2/3)
};

static uint8_t loadErrCode;
static int64_t *BPM2TickSamplesTab64; // 32.32fp (PT-range BPM, 32..255. 256 = Vblank mode)
static ptSmp_t *samplePtrs[PT_SAMPLES];

// globals
volatile bool Replayer_ModuleLoaded, Replayer_SongPlaying;
uint8_t Replayer_PlayMode;
ptSmp_t samples[PT_SAMPLES];
ptMod_t module;
ptChn_t channel[PT_CHANNELS];
note_t *patternData;
uint32_t sampleOffsets, moduleOffset, channelOffset, patternDataOffset;
uint32_t Replayer_PlaybackSeconds, Replayer_PlaybackSecondsFrac;

void Replayer_UpdateRepeats(ptSmp_t *s)
{
	const uint8_t testSample = 1 + ui.currSample;
	const uint32_t loopStartInBytes = (uint32_t)s->loopStart << 1;
	paula_t *p = paula;
	ptChn_t *ch = channel;
	int32_t i;
	
	_disable(); // turn off PIT audio mixing interrupt
	
	for (i = 0; i < PAULA_VOICES; i++, p++, ch++)
	{
		if (ch->n_samplenum == testSample)
		{
			// PT2.3D bugfix: update loops for replayer as well
			ch->n_loopstart = ch->n_start + loopStartInBytes;
			ch->n_replen = s->loopLength;
			
			Paula_SetDat(p, ch->n_loopstart);
			Paula_SetLen(p, ch->n_replen);
		}
	}

	_enable(); // turn back on PIT audio mixing interrupt
}

void Replayer_Update(uint32_t row, uint32_t order)
{
	module.row = (uint8_t)row;
	ui.redrawPatternFlag = true;

	if (module.order != order)
	{
		module.order = (uint8_t)order;
		module.pattern = module.orderList[order & 127];	
		ui.redrawPositionFlag = true;
	}
}

void Replayer_CalculateModSize(void)
{
	uint32_t i, totalModuleSize, numPatterns;
	ptSmp_t *s;

	totalModuleSize = 1084; // base size (includes no patterns/sample datas)

	// count number of patterns and add to final size
	numPatterns = 0;
	for (i = 0; i < PT_MAX_ORDERS; i++)
	{
		if (module.orderList[i] > numPatterns)
			numPatterns = module.orderList[i];
	}
	totalModuleSize += (numPatterns + 1) << 10; // *1024

	// add all sample lengths to final size
	s = samples;
	for (i = 0; i < PT_SAMPLES; i++, s++)
		totalModuleSize += (uint32_t)s->length << 1; // *2

	module.modSize = totalModuleSize;
	ui.redrawModSizeFlag = true;
}

void Replayer_SetBPM(uint32_t bpm)
{
	if (bpm < PT_MIN_BPM)
		bpm = PT_MIN_BPM;

	module.BPM = (uint8_t)bpm;

	if (module.useVblankTempo)
	{
		// vblank mode (~49.92Hz)
		Audio_SetSamplesPerTick(BPM2TickSamplesTab64[(PT_MAX_BPM-PT_MIN_BPM)+1]);
		ui.redrawBPMFlag = true;
		return;
	}

	Audio_SetSamplesPerTick(BPM2TickSamplesTab64[bpm-PT_MIN_BPM]);
	ui.redrawBPMFlag = true;
}

bool Replayer_Init(void)
{
	int8_t i;
	ptSmp_t *s;

	ptChn_t *ch = channel;
	for (i = 0; i < PT_CHANNELS; i++, ch++)
	{
		ch->n_chanindex = i;
		ch->n_dmabit = 1 << i;
	}

	patternData = (note_t *)malloc(PT_MAX_PATTERNS * 1024);
	if (patternData == NULL)
		return false;
		
	ASM_FastFill32(patternData, 0x00000000, (PT_MAX_PATTERNS * 1024) / 4);

	module.speed = 6;
	Replayer_SetBPM(125);

	s = samples;
	for (i = 0; i < PT_SAMPLES; i++, s++)
	{
		samplePtrs[i] = s; // for areplay.asm
		s->loopLength = 1;
	}

	sampleOffsets = (uint32_t)samplePtrs;
	moduleOffset = (uint32_t)&module;
	channelOffset = (uint32_t)channel;
	patternDataOffset = (uint32_t)patternData;

	return true;
}

void Replayer_Close(void)
{
	if (patternData != NULL)
	{
		free(patternData);
		patternData = NULL;
	}
}

static double ProTrackerBPM2Hz(uint8_t bpm)
{
	int32_t CIAPeriod;
	double dHz;

	if (bpm < PT_MIN_BPM)
		bpm = PT_MIN_BPM;

	CIAPeriod = 1773447 / bpm; // exactly how PT calculates BPM -> CIA period
	return (double)CIA_PAL_CLK / (CIAPeriod+1); // +1, CIA triggers on underflow
}

bool Replayer_InitBPMTable(void)
{
	int32_t bpm;
	double dSamplesPerTick;

	BPM2TickSamplesTab64 = (int64_t *)malloc(((PT_MAX_BPM-PT_MIN_BPM)+1+1) * sizeof (int64_t));
	if (BPM2TickSamplesTab64 == NULL)
		return false;

	// generate BPM -> "samples per tick" table (covers ProTracker BPM range)
	for (bpm = PT_MIN_BPM; bpm <= PT_MAX_BPM; bpm++)
	{
		dSamplesPerTick = audio.dOutputFreq / ProTrackerBPM2Hz((uint8_t)bpm);
		BPM2TickSamplesTab64[bpm-PT_MIN_BPM] = (int64_t)((dSamplesPerTick * (UINT32_MAX+1.0)) + 0.5); // rounded 32.32fp
	}

	// vblank tempo mode (~49.92Hz)
	dSamplesPerTick = audio.dOutputFreq / AMIGA_PAL_VBLANK_HZ;
	BPM2TickSamplesTab64[(PT_MAX_BPM-PT_MIN_BPM)+1] = (int64_t)((dSamplesPerTick * (UINT32_MAX+1.0)) + 0.5); // rounded 32.32fp

	return true;
}

void Replayer_FreeBPMTable(void)
{
	if (BPM2TickSamplesTab64 != NULL)
	{
		free(BPM2TickSamplesTab64);
		BPM2TickSamplesTab64 = NULL;
	}
}

void Replayer_SetPos(uint8_t pos)
{
	_disable();
	Replayer_PrepareSetPos(pos);
	ui.redrawPatternFlag = true;
	ui.redrawPositionFlag = true;
	_enable();
}

void Replayer_Free(void)
{
	int32_t i;
	ptSmp_t *s;
	ptChn_t *ch;

	Replayer_SongPlaying = false;
	Paula_ClearVoices();

	_disable();

	s = samples;
	for (i = 0; i < PT_SAMPLES; i++, s++)
		Sampler_FreeSampleData(s);

	memset(samples, 0, sizeof (samples));
	memset(&module, 0, sizeof (module));
	memset(channel, 0, sizeof (channel));

	ch = channel;
	for (i = 0; i < PT_CHANNELS; i++, ch++)
	{
		ch->n_chanindex = i;
		ch->n_dmabit = 1 << i;
	}
	
	s = samples;
	for (i = 0; i < PT_SAMPLES; i++, s++)
		s->loopLength = 1;

	Replayer_ModuleLoaded = false;
	_enable();
}

bool Replayer_Load(const char *filename)
{
	FILE *f;
	uint32_t i, ID;
	uint8_t order[PT_MAX_ORDERS], numOrders, numPatterns;
	ptSmp_t *s;
	note_t *n;

	f = fopen(filename, "rb");
	if (f == NULL)
	{
		loadErrCode = LERR_FILE;
		return false;
	}

	// read magic ID and check it
	SEEK_TO(1080);
	READ_LONG(ID);
	if (ID != ID_FLT4 && ID != ID_MK64 && ID != ID_MK100)
		goto errNotAMod;

	// read order length and check it
	SEEK_TO(950);
	READ_BYTE(numOrders);
	if (numOrders == 0 || numOrders > 129)
		goto errNotAMod;

	// fixes a specific copy of beatwave.mod (XXX: Hackish?)
	if (numOrders == 129)
		numOrders = 128;

	// read order list and find number of patterns
	SEEK_TO(952);
	READ_BYTES(order, PT_MAX_ORDERS);

	numPatterns = 0;
	for (i = 0; i < PT_MAX_ORDERS; i++)
	{
		if (order[i] > numPatterns)
			numPatterns = order[i];
	}
	numPatterns++;

	/* Test if number of patterns is within normal range.
	** Exception: Non-Amiga trackers, but we don't support that.
	*/
	if (numPatterns > 100)
		goto errNotAMod;

	/* We assume that this is a compatible Amiga MOD at this point.
	** Clear module struct and start further loading.
	*/
	Replayer_ModuleLoaded = false;
	Replayer_Free();

	module.numPatterns = numPatterns;
	module.length = numOrders;

	// copy over order list
	memcpy(module.orderList, order, PT_MAX_ORDERS);

	// read song name
	SEEK_TO(0);
	READ_BYTES(module.name, 22); module.name[22] = '\0';

	// read sample headers and count total sample data size
	SEEK_TO(20);
	s = samples;
	for (i = 0; i < PT_SAMPLES; i++, s++)
	{
		READ_BYTES(s->name, 22); s->name[22] = '\0';
		READ_MOTOROLA_WORD(s->length);
		READ_BYTE(s->finetune); s->finetune &= 0xF;
		READ_BYTE(s->volume);
		READ_MOTOROLA_WORD(s->loopStart);
		READ_MOTOROLA_WORD(s->loopLength);

		// extra sanitation
		if ((uint8_t)s->volume > 64)
			s->volume = 64;

		// ProTracker does this
		if (s->loopStart+s->loopLength > s->length)
			s->loopStart >>= 1;

		// non-PT fix for illegal loops
		if (s->loopStart+s->loopLength > 1) // loop enabled?
		{
			if (s->loopStart >= s->length)
			{
				// disable loop
				s->loopStart = 0;
				s->loopLength = 1;
			}
			else if (s->loopStart+s->loopLength > s->length)
			{
				// limit overflown loop to sample end
				s->loopLength = s->length - s->loopStart;
			}
		}
		
		if (s->loopLength < 1)
		{
			s->loopStart = 0;
			s->loopLength = 1;
		}
	
		if (!Sampler_AllocateSampleData(s, s->length << 1))
			goto errOutOfMem;
	}

	SEEK_TO(1084);

	// read pattern data

	READ_BYTES(patternData, numPatterns << 10);
	Replayer_ConvertPatternData(patternData, numPatterns);

	// if we loaded less than 100 patterns, clear the remains
	if (numPatterns < PT_MAX_PATTERNS)
	{
		const uint32_t patternsToClear = PT_MAX_PATTERNS - numPatterns;
		ASM_FastFill32(&patternData[numPatterns << 8], 0x00000000, patternsToClear << (10-2));
	}
	
	// load sample data
	s = samples;
	for (i = 0; i < PT_SAMPLES; i++, s++)
	{
		if (s->length == 0)
			continue;

		// load sample data (ignore failure, some MODs have cut sample data at end of mod)
		fread(s->data, 1, s->length << 1, f);
		
		// fix infinite beep on non-looping samples (if first two bytes are not equal)
		if (s->loopStart+s->loopLength <= 1)
		{
			s->data[0] = 0;
			s->data[1] = 0;
		}
	}

	module.speed = PT_DEF_SPD;
	Replayer_SetBPM(PT_DEF_BPM);

	fclose(f);
	loadErrCode = LERR_SUCCESS;
	Replayer_ModuleLoaded = true;

	return true;

errNotAMod:
	fclose(f);
	loadErrCode = LERR_NOT_A_MOD;
	return false;

errIO:
	fclose(f);
	Replayer_Free();
	loadErrCode = LERR_IO;
	return false;

errOutOfMem:
	fclose(f);
	Replayer_Free();
	loadErrCode = LERR_OUT_OF_MEM;
	return false;
}

uint8_t Replayer_GetLoadError(void)
{
	return loadErrCode;
}

void Replayer_ShowLastError(void)
{
	switch (Replayer_GetLoadError())
	{
		case LERR_FILE:
			UI_Error("CAN'T OPEN FILE !");
		break;

		case LERR_IO:
			UI_Error("GENERAL I/O ERROR");
		break;

		case LERR_OUT_OF_MEM:
			UI_Error("OUT OF MEMORY !");
		break;

		case LERR_NOT_A_MOD:
			UI_Error("UNSUPPORTED MOD !");
		break;

		default: 
			UI_Error("UNKNOWN ERROR !");
		break;
	}
}

void Replayer_Play(uint8_t playMode, uint8_t startRow)
{
	Replayer_Stop();

	_disable();

	module.row = startRow;
	module.editPattern = module.pattern = module.orderList[module.order];
	module.tick = module.speed-1;

	Replayer_PreparePlay();
	Audio_ClearSampleCounter(); // immediately initiate a replayer tick

	Replayer_PlayMode = playMode;
	if (Replayer_PlayMode == PLAY_MODE_SONG)
		Replayer_PlaybackSeconds = Replayer_PlaybackSecondsFrac = 0; // clear playback time counter

	Replayer_SongPlaying = true;

	_enable();

	ui.redrawPatternFlag = true;
	Mouse_UpdateColor();
}

void Replayer_Stop(void)
{
	_disable();

	Paula_ClearVoices();
	Replayer_PrepareStop();

	Replayer_SongPlaying = false;
	Replayer_PlayMode = PLAY_MODE_IDLE;

	_enable();
}
