#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include "dma.h"

#define DMA_CHANNELS 8

typedef struct
{
	uint16_t length;
	void *buffer;
	uint8_t mode;
} dma_t;

static uint16_t dmaMemBlockSelector;
static bool initialized[DMA_CHANNELS], active[DMA_CHANNELS];

static const dmaports_t dmaPorts[8] =
{
	{ 0x00, 0x01, 0x87, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0x02, 0x03, 0x83, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0x04, 0x05, 0x81, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0x06, 0x07, 0x82, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0xC0, 0xC2, 0x8F, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC },
	{ 0xC4, 0xC6, 0x8B, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC },
	{ 0xC8, 0xCA, 0x89, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC },
	{ 0xCC, 0xCE, 0x8A, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC }
};

static dma_t dma[DMA_CHANNELS];

static void freeMemoryBlock(uint16_t blockSelector)
{
	union REGS r;
	r.w.ax = 0x0101;
	r.w.dx = blockSelector;
	int386(0x31, &r, &r);
}

bool DMA_Init(uint8_t channel, uint16_t len, uint8_t mode)
{
#define MAX_SELECTOR_TESTS 16

	union REGS r;
	bool memoryAllocated;
	void *buf;
	int16_t testedSelectors, i, oldBlockSelectors[MAX_SELECTOR_TESTS];

	if (channel >= DMA_CHANNELS || initialized[channel])
		return false;

	// allocate single-segmented DMA memory

	memoryAllocated = false;
	testedSelectors = 0;

	while (true)
	{
		// allocate memory
		r.w.ax = 0x0100;
		r.w.bx = (len + 15) >> 4;
		int386(0x31, &r, &r);

		if (r.x.cflag != 0) // out of memory
		{
			buf = NULL;
			break;
		}

		buf = (void *)(r.w.ax << 4);
		oldBlockSelectors[testedSelectors++] = dmaMemBlockSelector = r.w.dx;

		// check if this allocation region is single-segmented
		if (((FP_OFF(buf)+len+15) & 0xF0000) == (FP_OFF(buf) & 0xF0000))
		{
			// no segment crossing, lock this region for use
			r.w.ax = 0x0600;
			r.w.bx = FP_OFF(buf) >> 16;
			r.w.cx = FP_OFF(buf) & 0xFFFF;
			r.w.si = 0;
			r.w.di = len;
			int386(0x31, &r, &r);
	
			memoryAllocated = true;
			break;
		}
		
		if (testedSelectors == MAX_SELECTOR_TESTS)
			break;
	}

	if (memoryAllocated)
	{
		dma_t *d = &dma[channel];

		d->buffer = buf;
		d->mode = mode | (channel & 3);
		d->length = len;

		initialized[channel] = true;
		
		testedSelectors--; // don't free the DMA buffer's selector now
	}

	// free allocated memory we couldn't use
	for (i = 0; i < testedSelectors; i++)
		freeMemoryBlock(oldBlockSelectors[i]);

	return memoryAllocated;
}

void DMA_Stop(uint8_t channel)
{
	dmaports_t *p;

	if (channel >= DMA_CHANNELS || !active[channel])
		return;

	p = (dmaports_t *)&dmaPorts[channel];
	outp(p->mask, (channel & 3) | 4); // disable DMA channel	
	active[channel] = false;
}

bool DMA_Close(uint8_t channel)
{
	union REGS r;
	dma_t *d;

	if (channel >= DMA_CHANNELS || !initialized[channel])
		return false;

	d = &dma[channel];

	DMA_Stop(channel);
	
	// unlock memory region
	r.w.ax = 0x0601;
	r.w.bx = FP_OFF(d->buffer) >> 16;
	r.w.cx = FP_OFF(d->buffer) & 0xFFFF;
	r.w.si = 0;
	r.w.di = d->length;
	int386(0x31, &r, &r);

	freeMemoryBlock(dmaMemBlockSelector);

	initialized[channel] = false;
	return true;
}

bool DMA_Start(uint8_t channel, bool pauseIRQs)
{
	uint16_t page, addr, count;
	dmaports_t *p;
	dma_t *d;

	if (channel >= DMA_CHANNELS || !initialized[channel] || active[channel])
		return false;

	d = &dma[channel];
	p = (dmaports_t *)&dmaPorts[channel];

	page = FP_OFF(d->buffer) >> 16;
	addr = FP_OFF(d->buffer) & 0xFFFF;

	count = d->length;
	if (channel >= 4) // 16-bit DMA
	{
		addr = ((page & 1) << 15) | (addr >> 1);
		count >>= 1;
	}

	count--;

	if (pauseIRQs)
		_disable();

	outp(p->mask, (channel & 3) | 4); // disable DMA channel
	outp(p->mode, d->mode);
	outp(p->flipflop, 0); // reset flip-flop (prepare word access)
	outp(p->addr, (uint8_t)addr);
	outp(p->addr, addr >> 8);
	outp(p->page, page);
	outp(p->flipflop, 0); // reset flip-flop (again)
	outp(p->count, (uint8_t)count);
	outp(p->count, count >> 8);
	outp(p->mask, channel & 3); // enable DMA channel

	if (pauseIRQs)
		_enable();

	active[channel] = true;
	return true;
}

void *DMA_GetBuffer(uint8_t channel)
{
	if (channel >= DMA_CHANNELS || !initialized[channel])
		return NULL;

	return dma[channel].buffer;
}

uint16_t DMA_GetPosition(uint8_t channel)
{
	uint16_t output, dmaCountPort, dmaFlipflopPort;

	if (channel >= DMA_CHANNELS || !initialized[channel])
		return 0;

	// based on code from Fasttracker II

	dmaCountPort = dmaPorts[channel].count;
	dmaFlipflopPort = dmaPorts[channel].flipflop;
	
	__asm
	{
		push eax
		push ebx
		push ecx
		push edx
		push edi
		// --------------------------------------------
		// Read DMA pos until the difference between two reads
		// is < 32, this is because of memory managers, etc.
		// --------------------------------------------
		mov bx,dmaCountPort
		mov cx,dmaFlipflopPort
		// ------------
	dmaPosLoop:
		mov dx,cx
		xor al,al
		out dx,al // reset flip-flop (prepare word access)
		// ------------
		mov dx,bx
		in al,dx
		mov ah,al
		in al,dx
		xchg al,ah
		mov di,ax
		// ------------
		mov dx,cx
		xor al,al
		out dx,al // reset flip-flop (again)
		// ------------
		mov dx,bx
		in al,dx
		mov ah,al
		in al,dx
		xchg al,ah
		// ------------
		sub di,ax
		jns notSigned
		neg di
	notSigned:
		// ------------
		cmp di,32 // difference < 32?
		jnb dmaPosLoop // nope, go again
		// ------------
		cmp byte ptr channel,4
		jb no16Bit
		add ax,ax
	no16Bit:
		mov output,ax
		pop edi
		pop edx
		pop ecx
		pop ebx
		pop eax
	}
	
	return output;
}
