#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include "helpers.h"
#include "hardware/keyb.h"
#include "hardware/mouse.h"
#include "hardware/video.h"
#include "hardware/dma.h"
#include "hardware/pit.h"
#include "audio/drivers/sb.h"
#include "audio/audio.h"
#include "audio/replayer.h"
#include "audio/paula.h"
#include "gfx/gfx.h"
#include "gfx/font.h"
#include "gfx/drawcode.h"
#include "dir.h"
#include "ui.h"
#include "scopes.h"
#include "sampler.h"
#include "pattern.h"

#define ARG(i, s) !stricmp(argv[i], s)
#define ARGN(i, s, n) !strnicmp(argv[i], s, n)

static bool noSound = false;

static void __interrupt (*oldInt36Vector)(void);
static void usage(void);
static bool handleArguments(int argc, char *argv[]);
static void disableDOSCriticalErrorHandler(void);
static void resetDOSCriticalErrorHandler(void);
static void exitCleanup(void);
static void InitCommonData(void);

static void __interrupt int36Vector(void) { }; // do nothing

volatile bool trackerRunning = false; // global

uint32_t *mouseStructPtr, *audioStructPtr; // for amouse.asm and apitint.asm

int main(int argc, char *argv[])
{
	InitCommonData();

	if (handleArguments(argc, argv))
		return 1; // usage was shown, quit now...

	if (!Video_IsColorVGA())
	{
		printf("ERROR: PT4DOS requires a VGA adapter connected to a color monitor.\n");
		return 1;
	}

	Video_GetVideoMode();
	if (!Mouse_Initialize()) // error message is printed in Mouse_Initialize()
		return 1;

	Audio_Detect(noSound);

	if (!Audio_Init()) // error message is printed in Audio_Init()
	{
		Audio_Close();
		Mouse_Close();
		return 1;
	}
	
	if (!Paula_Init())
	{
		Paula_Free();
		Audio_Close();
		Mouse_Close();
		printf("Error: Out of memory!\n");
		return 1;
	}

	if (!Keyb_Initialize())
	{
		exitCleanup();
		printf("ERROR: Couldn't replace interrupt 1 vector (keyboard)!\n");
		return 1;
	}

	if (!UI_CreateScopeTable() || !DrawCode_MakeCode() || !Sampler_Init())
	{
		exitCleanup();
		printf("Error: Out of memory!\n");
		return 1;
	}

	disableDOSCriticalErrorHandler();
	Dir_Init();
	Video_SetNewVideoMode();
	Video_WaitForVblank();
	Video_MeasureVblankPitTicks();
	PIT_StartVblankInterrupt();

	if (!Replayer_Init())
	{
		exitCleanup();
		printf("ERROR: Out of memory!\n");
		return 1;
	}

	ui.editSkip = 1;
	module.length = 1;
	module.speed = 6;
	module.BPM = 125;
	Replayer_SetBPM(125);
	Replayer_ModuleLoaded = true;
	Replayer_SongPlaying = false;
	Replayer_CalculateModSize();

	UI_DrawTopScreen();
	UI_DrawMiddleScreen();
	UI_DrawBottomScreen();
	Pattern_SetCursorPos(0);

	if (!Audio_StartDMA())
	{
		exitCleanup();
		printf("ERROR: Couldn't start audio DMA!\n");
		return 1;
	}

	UI_StatusAllRight();
	Mouse_Update();
	trackerRunning = true;

	if (argc > 1 && argv[1][0] != '/')
	{
		UI_SetStatusText("LOADING...", NO_CARRY);
		Mouse_SetDiskAccessColor();

		if (!Replayer_Load(argv[1]))
		{
			Replayer_ShowLastError();
		}
		else
		{
			Replayer_Play(PLAY_MODE_SONG, 0);
			Replayer_CalculateModSize();

			UI_DrawSampleInfo();
			UI_DrawSongName();
			UI_DrawSampleName();
			ui.redrawPositionFlag = true;
			ui.redrawBPMFlag = true;
			ui.redrawPatternFlag = true;
			ui.redrawModSizeFlag = true;
			
			UI_StatusAllRight();
		}
		
		Mouse_UpdateColor();
	}

	while (trackerRunning)
	{
		Video_WaitForVblank();
		
		Sampler_EraseOldSamplingLine();
		
		Keyb_HandleInput();	
		Mouse_HandleButtonRepeat();
		Mouse_HandleInput();
		UI_AmigaVblankHandler();
		UI_HandleUpdates();

		Sampler_WriteSamplingLine();
	}

	exitCleanup();
	
	printf("PT4DOS alpha - compiled at %s\n", __DATE__);
	printf("Visit https://16-bits.org for more tracker related stuff.\n");	
	
	return 0;
}

static void disableDOSCriticalErrorHandler(void)
{
	union REGS r;
	struct SREGS sr;

	// get old int36 vector
	segread(&sr);
	r.w.ax = 0x3524;
	int386x(0x21, &r, &r, &sr);
	oldInt36Vector = MK_FP(sr.es, r.w.bx);

	// set new int36 vector
	segread(&sr);
	r.w.ax = 0x2524;
	r.x.edx = FP_OFF(int36Vector);
	sr.ds = FP_SEG(int36Vector);
	int386x(0x21, &r, &r, &sr);
}

static void resetDOSCriticalErrorHandler(void)
{
	union REGS r;
	struct SREGS sr;

	segread(&sr);
	r.w.ax = 0x2524;
	r.x.edx = FP_OFF(oldInt36Vector);
	sr.ds = FP_SEG(oldInt36Vector);
	int386x(0x21, &r, &r, &sr);
}

static void exitCleanup(void)
{
	if (showRasterBars)
		Video_SetBackgroundColor(0, 0, 0);

	PIT_StopVblankInterrupt();
	Audio_Close();
	Mouse_Close();
	Replayer_Free();
	Paula_Free();
	Sampler_Free();
	Dir_Free();
	Dir_Reset();
	Keyb_Close();
	UI_FreeScopeTable();
	Replayer_Close();
	DrawCode_FreeCode();
	Video_ResetVideoMode();
	UI_PrintQuitMessage(); // debugging

	resetDOSCriticalErrorHandler();
}

static void InitCommonData(void)
{
	memset(&audio,  0, sizeof (audio));
	memset(&ui,     0, sizeof (ui));
	memset(&module, 0, sizeof (module));
	memset(&mouse,  0, sizeof (mouse));
	memset(&keyb,   0, sizeof (keyb));
	
	audioStructPtr = (uint32_t *)&audio;
	mouseStructPtr = (uint32_t *)&mouse;
	
	audio.defStereoSep = DEF_STEREO_SEP;

	ui.f6Pos = 0;
	ui.f7Pos = 16;
	ui.f8Pos = 32;
	ui.f9Pos = 48;
	ui.f10Pos = 63;
}

static void usage(void)
{
	printf("Usage:\n");
	printf("  pt4dos <input_module> [/nosound] [/a500] [-stereosep=num]\n");
	printf("\n");
	printf("  Options:\n");
	printf("    input_module      Optional filename for a module to play\n");
	printf("    /nosound          Don't attempt to initialize a Sound Blaster card\n");
	printf("    /a500             Start tracker with Amiga 500 filter enabled\n");
	printf("    -stereosep=num    Set stereo separation in percentage (0..100).\n");
	printf("                      Default is %d. Affects Sound Blaster 16 only.\n", DEF_STEREO_SEP);
}

static bool handleArguments(int argc, char *argv[])
{
	if (argc > 1)
	{
		int32_t i;

		if (argc == 2 && ARG(1, "/?"))
		{
			usage();
			return true;
		}
		
		for (i = 1; i < argc; i++)
		{
			if (ARG(i, "/nosound"))
			{
				noSound = true;
			}
			else if (ARG(i, "/a500"))
			{
				audioAmiga500 = true;
			}
			else if (ARGN(i, "-stereosep=", 11))
			{
				const int32_t percentage = atoi(&argv[i][11]);
				audio.defStereoSep = (int8_t)CLAMP(percentage, 0, 100);
			}
		}

	}
	
	return false;
}
