#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <conio.h>
#include "helpers.h"
#include "hardware/mouse.h"
#include "hardware/video.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "ui.h"
#include "asm.h"
#include "sampler.h"
#include "pattern.h"
#include "scopes.h" // SCOPE_W
#include "audio/paula.h"
#include "audio/replayer.h"

#define SAMPLE_AREA_X 3
#define SAMPLE_AREA_Y 138
#define SAMPLE_AREA_H 48
#define SAMPLE_AREA_W 314
#define SAMPLE_AREA_Y_PLANEOFFSET (SAMPLE_AREA_Y*(SCREEN_W/8))
#define SAMPLE_AREA_Y_CENTER (SAMPLE_AREA_Y+23)
#define SAMPLE_AREA_BYTES (SAMPLE_AREA_H * (SCREEN_W/8))

sampler_t sampler; // globalized
uint8_t *wavePlaneBuffer;

static uint8_t waveScaleTable[256], *wavePlaneBufferOrigPtr;
static int32_t lastLoopPin1x = -1, lastLoopPin2x = -1, oldSamplingLineXPos;

// at bottom of this file
static const uint32_t emptySample[SAMPLE_AREA_BYTES/4];
static const uint16_t leftLoopPinGfx[48*8], rightLoopPinGfx[48*9];

bool Sampler_Init(void)
{
	int32_t i;
	for (i = 0; i < 256; i++)
		waveScaleTable[i] = ((SAMPLE_AREA_H/2)-1) - (((int8_t)i * SAMPLE_AREA_H) >> 8);

	wavePlaneBufferOrigPtr = (uint8_t *)malloc((SAMPLE_AREA_BYTES * sizeof (uint32_t)) + 4);
	if (wavePlaneBufferOrigPtr == NULL)
		return false;

	wavePlaneBuffer = (uint8_t *)(((uint32_t)wavePlaneBufferOrigPtr + 3) & ~3); // dword aligned (just in case)

	return true;
}

void Sampler_Free(void)
{
	if (wavePlaneBufferOrigPtr != NULL)
	{
		free(wavePlaneBufferOrigPtr);
		wavePlaneBufferOrigPtr = NULL;
	}
}

bool Sampler_AllocateSampleData(ptSmp_t *s, int32_t lengthInBytes)
{
	// some padding is needed for the quadrascope
	s->data = (int8_t *)malloc(lengthInBytes + SCOPE_W); 
	if (s->data == NULL)
		return false;

	memset(&s->data[lengthInBytes], 0, SCOPE_W); // clear the padding
	return true;
}

void Sampler_FreeSampleData(ptSmp_t *s)
{
	if (s->data != NULL)
	{
		free(s->data);
		s->data = NULL;	
	}
	
	s->length = 0;
}

void Sampler_SetLoopStartFromX(ptSmp_t *s, int32_t x, int32_t oldLoopEnd)
{
	volatile int32_t loopStart, loopLength;

	x = CLAMP(x, 0, 314);

	x = (sampler.samOffset + ((x * sampler.samDisplay) / 314)) >> 1;
	if (x >= oldLoopEnd)
		x = oldLoopEnd-1;

	loopStart = s->loopStart;
	loopLength = s->loopLength;

	loopLength += loopStart - x;
	loopStart = x;

	if (loopStart+loopLength > s->length)
		loopLength = s->length - loopStart;

	if (loopLength < 1)
		loopLength = 1;

	if (loopStart != s->loopStart || loopLength != s->loopLength)
	{
		s->loopStart = loopStart;
		s->loopLength = loopLength;

		if (!ui.diskOpShown)
		{
			Video_SelectWritePlane(2);
			Font_PrintHexWord(64, 80, s->loopStart << 1);
			Font_PrintHexWord(64, 91, s->loopLength << 1);
		}
		
		Replayer_UpdateRepeats(s);	
		Sampler_RedrawWaveform();
	}
}

void Sampler_SetLoopEndFromX(ptSmp_t *s, int32_t x)
{
	volatile int32_t loopLength;
	
	x = CLAMP(x-3, 0, 314);

	x = (sampler.samOffset + ((x * sampler.samDisplay) / 314)) >> 1;
	if (x < s->loopStart)
		x = s->loopStart;

	loopLength = x - s->loopStart;
	if (loopLength < 1)
		loopLength = 1;	

	if (loopLength != s->loopLength)
	{
		s->loopLength = loopLength;

		if (!ui.diskOpShown)
		{
			Video_SelectWritePlane(2);
			Font_PrintHexWord(64, 91, s->loopLength << 1);
		}
		
		Replayer_UpdateRepeats(s);
		Sampler_RedrawWaveform();
	}
}

void Sampler_EraseOldSamplingLine(void)
{
	if (ui.samplerScreenShown)
	{
		if (oldSamplingLineXPos >= SAMPLE_AREA_X && oldSamplingLineXPos < SAMPLE_AREA_X+SAMPLE_AREA_W)
			GFX_InvertSamplerLine(oldSamplingLineXPos);
	}

	oldSamplingLineXPos = -1;
}


static int32_t Sampler_SamplePosToX(int32_t samplePos)
{
	if (sampler.samDisplay == 0 || samplePos < 0 || samplePos >= 131072)
		return 0;
		
	return SAMPLE_AREA_X + (((samplePos - sampler.samOffset) * SAMPLE_AREA_W) / sampler.samDisplay);
}

void Sampler_WriteSamplingLine(void)
{
	int8_t *smpStart, *smpEnd;
	int32_t i;
	uint32_t currSamplingPos;
	ptSmp_t *s;
	paula_t *p;

	if (!ui.samplerScreenShown || sampler.samDisplay <= 0 || oldSamplingLineXPos != -1)
		return;

	currSamplingPos = UINT32_MAX;

	s = &samples[ui.currSample];
	smpStart = s->data;
	smpEnd = s->data + (s->length << 1);

	p = paula;
	for (i = 0; i < PAULA_VOICES; i++, p++)
	{
		if (p->dmaActive && p->dataPtr >= smpStart && p->dataPtr < smpEnd)
			currSamplingPos = (uint32_t)(p->dataPtr - smpStart);
	}

	if (currSamplingPos != UINT32_MAX)
	{
		int32_t xPos = Sampler_SamplePosToX(currSamplingPos);
		if (xPos >= SAMPLE_AREA_X && xPos < SAMPLE_AREA_X+SAMPLE_AREA_W)
		{
			GFX_InvertSamplerLine(xPos);
			oldSamplingLineXPos = xPos;
		}
	}
}

void Sampler_MarkToOffset(void)
{
	if (sampler.markStart == 0)
		return;
		
	sampler.markStartOfs = sampler.samOffset + (((sampler.markStart - 3) * sampler.samDisplay) / SAMPLE_AREA_W);
	
	if (sampler.markEnd >= SAMPLE_AREA_W+2)
		sampler.markEndOfs = sampler.samOffset + sampler.samDisplay;
	else
		sampler.markEndOfs = sampler.samOffset + (((sampler.markEnd - 3) * sampler.samDisplay) / SAMPLE_AREA_W);
}

void Sampler_OffsetToMark(void)
{
	if (sampler.markStartOfs < 0)
	{
		sampler.markStart = 0;
		return;
	}
	
	if (sampler.samDisplay == 0)
	{
		sampler.markStart = sampler.markEnd = 3;
		return;
	}

	sampler.markStart = 3;
	if (sampler.markStartOfs >= sampler.samOffset)
	{
		sampler.markStart = 3 + ((sampler.markStartOfs - sampler.samOffset) * SAMPLE_AREA_W) / sampler.samDisplay;
		if (sampler.markStart > SAMPLE_AREA_W)
			sampler.markStart = 0;
	}
	
	sampler.markEnd = 3;
	if (sampler.markEndOfs < sampler.samOffset)
	{
		sampler.markStart = 0;
	}
	else
	{
		sampler.markEnd = 3 + ((sampler.markEndOfs - sampler.samOffset) * SAMPLE_AREA_W) / sampler.samDisplay;
		if (sampler.markEnd > SAMPLE_AREA_W-1)
			sampler.markEnd = SAMPLE_AREA_W-1;
	}
}

static void eraseLoopPin(uint16_t *src16, uint16_t *dst16)
{
__asm
{
	push eax
	push ecx
	push esi
	push edi
	mov esi,src16
	mov edi,dst16	
	mov ecx,SAMPLE_AREA_H/4
eloop:	mov ax,[esi+(0*(SCREEN_W/8))]
	mov [edi+(0*(SCREEN_W/8))],ax
	mov ax,[esi+(1*(SCREEN_W/8))]
	mov [edi+(1*(SCREEN_W/8))],ax
	mov ax,[esi+(2*(SCREEN_W/8))]
	mov [edi+(2*(SCREEN_W/8))],ax
	mov ax,[esi+(3*(SCREEN_W/8))]
	mov [edi+(3*(SCREEN_W/8))],ax
	add edi,(SCREEN_W/8)*4
	add esi,(SCREEN_W/8)*4
	dec ecx
	jnz eloop
	pop edi
	pop esi
	pop ecx
	pop eax
};
}

static void drawLoopPin(uint16_t *src16, uint16_t *dst16)
{
__asm
{
	push eax
	push ecx
	push edx
	push esi
	push edi
	mov esi,src16
	mov edi,dst16		
	mov ecx,SAMPLE_AREA_H
	mov dx,0x0300
dloop:	mov dl,0xCE
	mov ax,0x0204
	out dx,ax
	mov dl,0xC4
	mov ax,0x0402
	out dx,ax
	mov ax,[esi]
	or [edi],ax		
	mov dl,0xCE
	mov ax,0x0004
	out dx,ax
	mov dl,0xC4
	mov ax,0x0102
	out dx,ax
	mov ax,[esi]
	or [edi],ax			
	add edi,SCREEN_W/8
	add esi,2		
	dec ecx
	jnz dloop
	pop edi
	pop esi
	pop edx
	pop ecx
	pop eax
};
}

void Sampler_SetLoopSprites(void)
{
	ptSmp_t *s;
	uint8_t *srcPlaneOffset, *dstPlaneOffset;
	uint16_t *src16, *dst16;
	int32_t x;

	sampler.loopStartPos = sampler.loopEndPos = 0;

	if (!ui.samplerScreenShown)
		return;
		
	s = &samples[ui.currSample];
		
	// draw LOOP flag
	Video_SelectWritePlane(2);
	if (s->loopStart+s->loopLength > 1)
		Font_Print(288, 209, " ON");
	else
		Font_Print(288, 209, "OFF");

	srcPlaneOffset = GFX_GetSamplerBitmap() + (17 * (SCREEN_W/8));
	dstPlaneOffset = (uint8_t *)VGA_FRAMEBUFFER + SAMPLE_AREA_Y_PLANEOFFSET;
	Video_SetPlaneBitmask(1);

	// erase previous left loop pin from bitplane #1
	if (lastLoopPin1x > -1)
	{
		// prevent erasing one byte column too much from right edge
		if (lastLoopPin1x > SCREEN_W-16)
			lastLoopPin1x = SCREEN_W-16;

		lastLoopPin1x >>= 3;
		eraseLoopPin((uint16_t *)(srcPlaneOffset + lastLoopPin1x),
		             (uint16_t *)(dstPlaneOffset + lastLoopPin1x));
		lastLoopPin1x = -1;
	}

	// erase previous right loop pin from bitplane #1
	if (lastLoopPin2x > -1)
	{
		// prevent erasing one byte column too much from right edge
		if (lastLoopPin2x > SCREEN_W-16)
			lastLoopPin2x = SCREEN_W-16;

		lastLoopPin2x >>= 3;
		eraseLoopPin((uint16_t *)(srcPlaneOffset + lastLoopPin2x),
		             (uint16_t *)(dstPlaneOffset + lastLoopPin2x));
		lastLoopPin2x = -1;
	}

	if (s->loopStart+s->loopLength <= 1 | s->data == NULL || s->length == 0 || sampler.samDisplay == 0)
		return;

	// left loop pin	
	sampler.loopStartPos = Sampler_SamplePosToX(s->loopStart<<1)-3;
	if (sampler.loopStartPos >= 0 && sampler.loopStartPos < SAMPLE_AREA_X+SAMPLE_AREA_W)
	{
		src16 = (uint16_t *)&leftLoopPinGfx[(sampler.loopStartPos & 7) * 48];	
		dst16 = (uint16_t *)(dstPlaneOffset + (sampler.loopStartPos >> 3));
		drawLoopPin(src16, dst16);
		lastLoopPin1x = sampler.loopStartPos;
	}

	// right loop pin
	sampler.loopEndPos = Sampler_SamplePosToX((s->loopStart+s->loopLength)<<1);	
	if (sampler.loopEndPos >= SAMPLE_AREA_X && sampler.loopEndPos <= SAMPLE_AREA_X+SAMPLE_AREA_W)
	{
		src16 = (sampler.loopEndPos == 317) ? (uint16_t *)&rightLoopPinGfx[8 * 48] : (uint16_t *)&rightLoopPinGfx[(sampler.loopEndPos & 7) * 48];	
		dst16 = (uint16_t *)(dstPlaneOffset + (sampler.loopEndPos >> 3));
		drawLoopPin(src16, dst16);
		lastLoopPin2x = sampler.loopEndPos;
	}
}

void Sampler_RedrawWaveform(void)
{
	uint8_t *smpPtr;
	uint32_t markX1, markX2, samDisplay, smpPosDeltaHi, smpPosFrac;
	
	if (!ui.samplerScreenShown)
		return;
	
	Video_SelectWritePlane(2);

	if (sampler.samStart == NULL || sampler.samDisplay == 0)
	{
		// draw empty sample
		__asm
		{
			push eax
			push ebx
			push esi
			push edi
			lea esi,emptySample
			mov edi,VGA_FRAMEBUFFER + SAMPLE_AREA_Y_PLANEOFFSET
			mov ebx,(SAMPLE_AREA_BYTES/4)/8
		xloop:	mov eax,[esi+(0*4)]
			mov [edi+(0*4)],eax
			mov eax,[esi+(1*4)]
			mov [edi+(1*4)],eax
			mov eax,[esi+(2*4)]
			mov [edi+(2*4)],eax
			mov eax,[esi+(3*4)]
			mov [edi+(3*4)],eax
			mov eax,[esi+(4*4)]
			mov [edi+(4*4)],eax
			mov eax,[esi+(5*4)]
			mov [edi+(5*4)],eax
			mov eax,[esi+(6*4)]
			mov [edi+(6*4)],eax
			mov eax,[esi+(7*4)]
			mov [edi+(7*4)],eax
			add esi,4*8
			add edi,4*8
			dec ebx
			jnz xloop
			pop edi
			pop esi
			pop ebx
			pop eax
		}
		
		Sampler_SetLoopSprites();
		return;
	}

	markX1 = sampler.markStart;
	markX2 = sampler.markEnd;
	
	if (markX1 > markX2)
	{
		markX2 = sampler.markEnd;
		markX1 = sampler.markStart;
	}
		
	smpPtr = (uint8_t *)sampler.samStart + sampler.samOffset;
	samDisplay = sampler.samDisplay;
	smpPosFrac = 0;
	
	// TODO: Properly adjust smpPtr&smpPosFrac to prevent waveform
	// "wiggle" while scrolling.
	
	Video_SelectReadPlane(2); // needed for sampler line routine
	
	__asm
	{
		pushad
		
		// --------------------------------
		// clear sample bitplane buffer
		// --------------------------------
		mov edi,wavePlaneBuffer
		xor eax,eax
		mov ebx,(SAMPLE_AREA_BYTES/4)/8
	loop0:	mov [edi+(0*4)],eax
		mov [edi+(1*4)],eax
		mov [edi+(2*4)],eax
		mov [edi+(3*4)],eax
		mov [edi+(4*4)],eax
		mov [edi+(5*4)],eax
		mov [edi+(6*4)],eax
		mov [edi+(7*4)],eax
		add edi,4*8
		dec ebx
		jnz loop0

		
		// --------------------------------
		// draw waveform in sample bitplane buffer
		// --------------------------------

		// calculate sample position delta
		xor edx,edx 
		mov eax,samDisplay
		mov ebx,SAMPLE_AREA_W
		div ebx
		mov edi,eax	// edi = smpPosDeltaHi
		xor eax,eax
		div ebx		// eax = smpPosDeltaLo	
		// --------------------------------
		mov esi,smpPtr
		xor ebx,ebx
		xor edx,edx	
		// --------------------------------
		mov dl,[esi]
		mov bl,[waveScaleTable+edx] // ebx = y1
		// --------------------------------
		mov ecx,3 // ecx = x
	loop1:	add smpPosFrac,eax
		adc esi,edi
		mov dl,[esi]
		mov dl,[waveScaleTable+edx] // edx = y2
		call GFX_DrawSamplerLine
		mov ebx,edx // y1 = y2
		inc ecx
		cmp cx,(3+SAMPLE_AREA_W)-1
		jb loop1
		
		
		// --------------------------------
		// invert sample data range (if marked)
		// --------------------------------		
		mov eax,markX1
		test eax,eax
		jz noMark
		mov esi,markX2
		cmp esi,eax
		jg noSwap
		xchg esi,eax
	noSwap:	inc esi
		xor ebx,ebx
		mov ecx,SAMPLE_AREA_H
		mov edi,wavePlaneBuffer
		call GFX_InvertRect			
	noMark:
		// --------------------------------
		// copy sample bitplane buffer to VRAM
		// --------------------------------
		mov esi,wavePlaneBuffer
		mov edi,VGA_FRAMEBUFFER + SAMPLE_AREA_Y_PLANEOFFSET
		mov ebx,(SAMPLE_AREA_BYTES/4)/8
	loop2:	mov eax,[esi+(0*4)]
		mov [edi+(0*4)],eax
		mov eax,[esi+(1*4)]
		mov [edi+(1*4)],eax
		mov eax,[esi+(2*4)]
		mov [edi+(2*4)],eax
		mov eax,[esi+(3*4)]
		mov [edi+(3*4)],eax
		mov eax,[esi+(4*4)]
		mov [edi+(4*4)],eax
		mov eax,[esi+(5*4)]
		mov [edi+(5*4)],eax
		mov eax,[esi+(6*4)]
		mov [edi+(6*4)],eax
		mov eax,[esi+(7*4)]
		mov [edi+(7*4)],eax
		add esi,4*8
		add edi,4*8
		dec ebx
		jnz loop2
			
		popad
	};
	
	Sampler_SetLoopSprites();
}

void Sampler_DrawDragBar(void)
{
	if (sampler.samLength > 0 && sampler.samDisplay != sampler.samLength)
	{
		uint32_t dragStart, dragEnd;
		
		dragStart = 4 + (((uint32_t)sampler.samOffset * 311) / (uint32_t)sampler.samLength);
		dragEnd = 5 + (((uint32_t)(sampler.samDisplay + sampler.samOffset) * 311) / (uint32_t)sampler.samLength);
		
		sampler.dragStart = dragStart;
		sampler.dragEnd = dragEnd;

	__asm
	{
		pushad
	
		// --------------------------------
		// clear drag bar area in bitplane buffer
		// --------------------------------
		mov edi,wavePlaneBuffer
		xor eax,eax
		mov ebx,(4*(SCREEN_W/8))/4/8
	loop1:	mov [edi+(0*4)],eax
		mov [edi+(1*4)],eax
		mov [edi+(2*4)],eax
		mov [edi+(3*4)],eax
		mov [edi+(4*4)],eax
		mov [edi+(5*4)],eax
		mov [edi+(6*4)],eax
		mov [edi+(7*4)],eax
		add edi,4*8
		dec ebx
		jnz loop1	

		// --------------------------------
		// fill drag bar rectangle in bitplane buffer
		// --------------------------------
		mov eax,dragStart
		mov esi,dragEnd
		xor ebx,ebx
		mov ecx,4
		mov edi,wavePlaneBuffer
		call GFX_FillRect
		
		// --------------------------------
		// copy bitplane buffer to VRAM
		// --------------------------------	
		mov esi,wavePlaneBuffer
		mov edi,VGA_FRAMEBUFFER + (190 * (SCREEN_W/8))
		mov ebx,(4*(SCREEN_W/8))/4/8
	loop2:	mov eax,[esi+(0*4)]
		mov [edi+(0*4)],eax
		mov eax,[esi+(1*4)]
		mov [edi+(1*4)],eax
		mov eax,[esi+(2*4)]
		mov [edi+(2*4)],eax
		mov eax,[esi+(3*4)]
		mov [edi+(3*4)],eax
		mov eax,[esi+(4*4)]
		mov [edi+(4*4)],eax
		mov eax,[esi+(5*4)]
		mov [edi+(5*4)],eax
		mov eax,[esi+(6*4)]
		mov [edi+(6*4)],eax
		mov eax,[esi+(7*4)]
		mov [edi+(7*4)],eax
		add esi,4*8
		add edi,4*8
		dec ebx
		jnz loop2
	
		popad
	};
	}
	else
	{
		// no visible drag bar, erase whole area in VRAM
		__asm
		{
			push eax
			push ebx
			push edi
			mov edi,VGA_FRAMEBUFFER + (190 * (SCREEN_W/8))
			xor eax,eax
			mov ebx,(4*(SCREEN_W/8))/4/8
		loop1:	mov [edi+(0*4)],eax
			mov [edi+(1*4)],eax
			mov [edi+(2*4)],eax
			mov [edi+(3*4)],eax
			mov [edi+(4*4)],eax
			mov [edi+(5*4)],eax
			mov [edi+(6*4)],eax
			mov [edi+(7*4)],eax
			add edi,4*8
			dec ebx
			jnz loop1	
			pop edi
			pop ebx
			pop eax
		};
	}
}

static void drawDisplayLength(void)
{
	char *txtPtr, dispString[6+1];

	if (sampler.samDisplay <= 0)
	{
		txtPtr = "     0";
	}
	else if (sampler.samDisplay > 131070) // shouldn't be possible, but just in case
	{
		txtPtr = "131070";
	}
	else
	{
		const uint32_t value = sampler.samDisplay;
		txtPtr = dispString;
	__asm
	{
		push eax
		push ebx
		push edx
		push edi

		lea edi,dispString
		mov byte ptr [edi+6],0 // NUL-terminator
		mov ebx,10 // divisor
		mov eax,value

		// digit 6
		xor edx,edx
		div ebx
		add dl,'0'
		mov [edi+5],dl
		test eax,eax
		jz digit6

		// digit 5
		xor edx,edx
		div ebx
		add dl,'0'
		mov [edi+4],dl
		test eax,eax
		jz digit5

		// digit 4
		xor edx,edx
		div ebx
		add dl,'0'
		mov [edi+3],dl
		test eax,eax
		jz digit4

		// digit 3
		xor edx,edx
		div ebx
		add dl,'0'
		mov [edi+2],dl
		test eax,eax
		jz digit3

		// digit 2
		xor edx,edx
		div ebx
		add dl,'0'
		mov [edi+1],dl
		test eax,eax
		jz digit2

		// digit 1
		xor edx,edx
		div ebx
		add dl,'0'
		mov [edi],dl
		jmp done
		
digit6:		mov [edi+4],' '
digit5:		mov [edi+3],' '
digit4:		mov [edi+2],' '
digit3:		mov [edi+1],' '
digit2:		mov [edi+0],' '
	
done:		pop edi
		pop edx
		pop ebx
		pop eax
	};	
	}
	
	Font_Print(264, 198, txtPtr);
}

void Sampler_ShowSampleInfo(void)
{
	UI_DrawSampleInfo();
	UI_DrawSampleName();
}

void Sampler_DisplaySample(void)
{
	ptSmp_t *s;
	
	if (!ui.samplerScreenShown)
		return;

	// clear sample marking
	sampler.markStartOfs = -1;
	sampler.markStart = 0;
		
	s = &samples[ui.currSample];
	sampler.samStart = s->data;
	sampler.samOffset = 0;
	sampler.samLength = (uint32_t)s->length << 1;
	sampler.samDisplay = sampler.samLength;

	Sampler_RedrawWaveform();
	
	Video_SelectWritePlane(2);
	Sampler_DrawDragBar();
	drawDisplayLength();
}

void Sampler_Toggle(void)
{
	ui.samplerScreenShown ^= 1;

	if (ui.samplerScreenShown)
	{
		lastLoopPin1x = lastLoopPin2x = -1;
		sampler.markStartOfs = -1;
		sampler.markStart = 0;
					
		GFX_DrawSamplerScreenBitmap();
		UI_DrawModSize();
		UI_DrawTempoMode();
		UI_DrawStatusText();
		UI_HandleUpdates();
		
		// for loop pins
		Video_SetPalColor(PAL_PATCURSOR,0, 63, 0);
		Video_SetPalColor(PAL_PATTXT,   0, 63, 0);
		Video_WaitForVblank();
	}
	else
	{
		GFX_DrawPatternEditorBitmap();
		UI_DrawModSize();
		UI_DrawTempoMode();
		UI_DrawSongTempo();
		UI_DrawStatusText();
		Pattern_ClearCache();
		ui.redrawPatternFlag = true;

		// restore palette entries
		Video_SetPalColor(PAL_PATCURSOR,
			origTrackerPal[PAL_PATCURSOR].r,
			origTrackerPal[PAL_PATCURSOR].g,
			origTrackerPal[PAL_PATCURSOR].b);
			
		Video_SetPalColor(PAL_PATTXT,
			origTrackerPal[PAL_PATTXT].r,
			origTrackerPal[PAL_PATTXT].g,
			origTrackerPal[PAL_PATTXT].b);
			
		Video_WaitForVblank();
			
		Pattern_RedrawCursor();
		UI_HandleUpdates();
	}

	if (ui.samplerScreenShown)
	{
		Sampler_DisplaySample();
		
		// TODO: show these based on a variable
		Video_SelectWritePlane(2);
		Font_Print(288, 220, "C-3"); // sample play note
		Font_Print(288, 231, "C-3"); // sample resample note
	}
}

static const uint32_t emptySample[SAMPLE_AREA_BYTES/4] =
{
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xFFFFFF1F,0xFFFFFFFF,
	0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,
	0xFFFFFFFF,0xFFFFFFFF,0xF8FFFFFF,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

static const uint16_t leftLoopPinGfx[48*8] =
{
	0x00F0,0x00F0,0x00F0,0x00F0,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0070,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x0078,0x0078,0x0078,0x0078,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0038,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x003C,0x003C,0x003C,0x003C,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x001C,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x001E,0x001E,0x001E,0x001E,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x000E,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0x000F,0x000F,0x000F,0x000F,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0007,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,
	0x8007,0x8007,0x8007,0x8007,0x8000,0x8000,0x8000,0x8000,
	0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,
	0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8003,
	0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,
	0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,
	0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,0x8000,
	0xC003,0xC003,0xC003,0xC003,0x4000,0x4000,0x4000,0x4000,
	0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,
	0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0xC001,
	0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,
	0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,
	0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,0x4000,
	0xE001,0xE001,0xE001,0xE001,0x2000,0x2000,0x2000,0x2000,
	0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,
	0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0xE000,
	0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,
	0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,
	0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000,0x2000
};

static const uint16_t rightLoopPinGfx[48*9] =
{
	0x00F0,0x00F0,0x00F0,0x00F0,0x0080,0x0080,0x0080,0x0080,
	0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,
	0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x00E0,
	0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,
	0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,
	0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,
	0x0078,0x0078,0x0078,0x0078,0x0040,0x0040,0x0040,0x0040,
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0070,
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,
	0x003C,0x003C,0x003C,0x003C,0x0020,0x0020,0x0020,0x0020,
	0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,
	0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0038,
	0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,
	0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,
	0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,0x0020,
	0x001E,0x001E,0x001E,0x001E,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x001C,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,
	0x000F,0x000F,0x000F,0x000F,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x000E,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,
	0x8007,0x8007,0x8007,0x8007,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0007,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0xC003,0xC003,0xC003,0xC003,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x8003,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,0x0002,
	0xE001,0xE001,0xE001,0xE001,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0xC001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,
	0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,0x0001,

	// specal edge case
	0x0007,0x0007,0x0007,0x0007,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0007,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,
	0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004,0x0004
};
