#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include "helpers.h"
#include "audio/audio.h"
#include "audio/paula.h"
#include "audio/replayer.h"
#include "hardware/keyb.h"
#include "hardware/mouse.h"
#include "hardware/video.h"
#include "gfx/gfx.h"
#include "gfx/font.h"
#include "ui.h"
#include "scopes.h"
#include "diskop.h"
#include "pattern.h"
#include "sampler.h"

#define STATUSTEXT_MAX_CHARS 17

static const char *ftuneStrTab[16] =
{
	" 0", "+1", "+2", "+3",
	"+4", "+5", "+6", "+7",
	"-8", "-7", "-6", "-5",
	"-4", "-3", "-2", "-1"
};

static char statusText[STATUSTEXT_MAX_CHARS+1], oldStatusText[STATUSTEXT_MAX_CHARS+1];
static int8_t oldPattern = -1, oldRow = -1;
static int16_t statusTextCounter, *ScopeTablePtr;
static int32_t statusTextLength;
static uint32_t oldPlaybackSeconds = UINT32_MAX;

char editSkipMsg[12+1] = "EDITSKIP = 0";
char endMsg[128] = {'\0'}; // to be printed after program exit, if first byte is not NUL

volatile bool showRasterBars;
ui_t ui;
int16_t *ScopeTable;

const char HexTab[16] = // globalized
{
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
};

void UI_SetPrevStatusMessage(void)
{
	if (oldStatusText != NULL)
		UI_SetStatusText(oldStatusText, NO_CARRY);
}

void UI_PrintQuitMessage(void)
{
	if (endMsg[0] != '\0')
	{
		printf(endMsg);
		printf("\n");
	}
}

void UI_AmigaVblank(void)
{
	/* TODO in here:
	** - Handle pattern cursor left/right
	** - Handle text cursor input + left/right
	*/
}

void UI_ShowTemporaryStatusText(const char *text, uint16_t duration)
{
	UI_SetStatusText(text, NO_CARRY);
	statusTextCounter = duration;
}

void UI_Error(const char *text)
{
	UI_ShowTemporaryStatusText(text, ERR_DURATION);
	Mouse_Error();
}

void UI_HandleUpdates(void)
{
	if (!ui.diskOpShown && !ui.setupShown)
	{
		if (!ui.askBoxShown)
			Scopes_Draw();

		if (ui.redrawPositionFlag)
		{
			ui.redrawPositionFlag = false;
			UI_DrawPosInfo();
		}
		
		if (ui.redrawTempoModeFlag)
		{
			ui.redrawTempoModeFlag = false;
			UI_DrawTempoMode();
		}

		if (ui.redrawBPMFlag)
		{
			ui.redrawBPMFlag = false;
			
			if (!ui.samplerScreenShown)
				UI_DrawSongTempo();
		}
	}

	UI_DrawPlaybackTimer();

	if (!ui.samplerScreenShown)
	{
		if (ui.redrawModSizeFlag)
		{
			ui.redrawModSizeFlag = false;
			UI_DrawModSize();
		}

		if (ui.redrawPatternFlag || module.pattern != oldPattern)
			UI_DrawEditPattern();

		if (ui.redrawPatternFlag || (module.pattern != oldPattern || module.row != oldRow))
		{
			Pattern_Draw(module.pattern);

			ui.redrawPatternFlag = false;
			oldPattern = module.pattern;
			oldRow = module.row;
		}
	}

	// handle temporary status text
	if (statusTextCounter > 0)
	{
		statusTextCounter--;
		if (statusTextCounter <= 0)
			UI_SetPrevStatusMessage();
	}
}

bool AskBox(const char *text)
{
	bool result;

	UI_SetStatusText(text, NO_CARRY);

	// fix potentially stuck states
	ui.markBlock = false;
	statusTextCounter = 0;
	mouse.leftButtonDown = false; 
	mouse.rightButtonDown = false;
	Keyb_EmptyBuffer();

	GFX_DrawAskBitmap();
	ui.askBoxShown = true;
	Mouse_UpdateColor();

	while (true)
	{
		Video_WaitForVblank();
			
		Sampler_EraseOldSamplingLine();

		if (Keyb_KeyPressed())
		{
			uint8_t scancode = Keyb_ReadKey();		
			if (scancode == KEY_Y || scancode == KEY_ENTER)
			{
				result = true;
				break;
			}
			else if (scancode == KEY_N || scancode == KEY_ESC)
			{
				result = false;
				break;
			}
		}

		if (mouse.leftButtonDown || mouse.rightButtonDown)
		{
			mouse.clickHandled = true;
			
			if (mouse.rightButtonDown)
			{
				result = false;
				break;
			}

			if (mouse.y >= 71 && mouse.y <= 81)
			{
				if (mouse.x >= 171 && mouse.x <= 196)
				{
					// YES
					result = true;
					break;
				}
				else if (mouse.x >= 234 && mouse.x <= 252)
				{
					// NO
					result = false;
					break;
				}
			}
		}

		UI_AmigaVblankHandler();
		UI_HandleUpdates();
		
		Sampler_WriteSamplingLine();
	}
	
	// done, remove ask box graphics

	if (ui.diskOpShown)
	{
		GFX_EraseAskBitmap();
		DiskOp_DrawFileList();
	}
	else if (ui.editOpShown)
	{
		/* TODO:
		** Check which Edit Op. screen we were on, or if
		** we can even trigger an askbox from edit op.
		** at all...
		*/
	}
	else
	{
		GFX_DrawScopesBitmap();
		Scopes_ResetStates();
		Scopes_Draw();
	}

	UI_StatusAllRight(); // XXX: Revert to previous status text?
	ui.askBoxShown = false;

	if (result == false)
		Mouse_Error();
	else
		Mouse_UpdateColor();

	return result;
}

bool UI_CreateScopeTable(void)
{
	uint32_t ScopeTableAddr;

	// table address must be a multiple of 512 for speed reasons
	ScopeTablePtr = (int16_t *)malloc(((32+1) * 256 * sizeof (int16_t)) + 512);
	if (ScopeTablePtr == NULL)
		return false;

	ScopeTableAddr = (uint32_t)ScopeTablePtr;
	ScopeTable = (int16_t *)((ScopeTableAddr + 511) & ~511); // aligned
	Scopes_GenerateTable();

	return true;
}

void UI_FreeScopeTable(void)
{
	if (ScopeTablePtr != NULL)
	{
		free(ScopeTablePtr);
		ScopeTablePtr = NULL;
	}
}

void UI_Debug(const char *text) // safe to use in interrupts/threads
{
	uint16_t oldPlaneMask, oldWriteMode, oldBitMask, oldLatchMode;
	
	// backup current VGA registers
	outpw(0x3C4, 0x0202); oldPlaneMask = inp(0x3C5);
	outpw(0x3CE, 0x0505); oldWriteMode = inp(0x3CF);
	outpw(0x3CE, 0x0808); oldBitMask   = inp(0x3CF);
	outpw(0x3CE, 0x0303); oldLatchMode = inp(0x3CF);
	
	// change VGA registers
	outpw(0x3C4, 0x0802); // 3rd plane
	outpw(0x3CE, 0x0005); // write mode #0
	outpw(0x3CE, 0xFF08); // bitmask
	outpw(0x3CE, 0x0003); // no data rotate

	Font_Print(0, 0, text);
	
	// set back VGA registers
	outpw(0x3C4, (oldPlaneMask << 8) | 0x02);
	outpw(0x3CE, (oldWriteMode << 8) | 0x05);
	outpw(0x3CE, (oldBitMask   << 8) | 0x08);
	outpw(0x3CE, (oldLatchMode << 8) | 0x03);
}

void UI_Debugf(const char *fmt, ...) // safe to use in interrupts/threads
{
	char strBuf[512+1];
	va_list args;

	// format the text string
	va_start(args, fmt);
	vsnprintf(strBuf, sizeof (strBuf)-1, fmt, args);
	va_end(args);
	
	UI_Debug(strBuf);
}

void UI_DrawStatusText(void)
{
	Video_SelectWritePlane(2);
	Font_Print(88, 127, statusText);

	// clear unused character blocks
	if (statusTextLength < STATUSTEXT_MAX_CHARS)
	{
		int32_t i, x = 88 + (statusTextLength << 3);
		for (i = statusTextLength; i < STATUSTEXT_MAX_CHARS; i++, x += 8)
			Font_PrintChar(x, 127, ' ');
	}
}

void UI_SetStatusText(const char *text, bool carry)
{
	int32_t textLength;

	if (carry)
	{
		strncpy(oldStatusText, text, STATUSTEXT_MAX_CHARS);
		oldStatusText[STATUSTEXT_MAX_CHARS] = '\0';
	}

	strncpy(statusText, text, STATUSTEXT_MAX_CHARS);
	statusText[STATUSTEXT_MAX_CHARS] = '\0';
	statusTextLength = strlen(statusText);

	UI_DrawStatusText();
}

void UI_StatusAllRight(void)
{
	UI_SetStatusText("ALL RIGHT", DO_CARRY);
}

void UI_DrawModSize(void)
{
	char txt[16];
	if (module.modSize <= 999999)
		sprintf(txt, "%6d", module.modSize);
	else
		sprintf(txt, " %4dK", module.modSize / 1000);

	Video_SelectWritePlane(2);
	Font_Print(264, 123, txt);
}

void UI_DrawTempoMode(void)
{
	Video_SelectWritePlane(2);
	Font_Print(288, 130, module.useVblankTempo ? "VBL" : "CIA");
}

void UI_DrawEditPattern(void)
{
	Video_SelectWritePlane(2);
	Font_Print2Dec(8, 127, module.editPattern);
}

void UI_DrawPosInfo(void)
{
	Video_SelectWritePlane(2);
	Font_Print3Dec(72, 3, module.order);
	Font_Print3Dec(72, 14, module.pattern);
}

void UI_DrawSongLength(void)
{
	Video_SelectWritePlane(2);
	Font_Print3Dec(72, 25, module.length);
}

void UI_DrawSampleInfo(void)
{
	ptSmp_t *s = &samples[ui.currSample];

	Video_SelectWritePlane(2);
	Font_Print(80, 36, ftuneStrTab[s->finetune & 0xF]);
	Font_PrintHexWord(64, 47, ui.currSample+1);
	Font_PrintHexWord(64, 58, s->volume);
	Font_PrintHexWord(64, 69, s->length<<1);
	Font_PrintHexWord(64, 80, s->loopStart<<1);
	Font_PrintHexWord(64, 91, s->loopLength<<1);
}

void UI_ResetPlaybackTimer(void)
{
	Replayer_PlaybackSeconds = Replayer_PlaybackSecondsFrac = 0;
}

void UI_DrawPlaybackTimer(void)
{
	volatile uint32_t playbackSeconds = Replayer_PlaybackSeconds;
	if (playbackSeconds != oldPlaybackSeconds)
	{
		uint16_t minutes, seconds;

		uint32_t secs = playbackSeconds;
		if (secs > 5999) // 99:59 (max for UI)
			secs = 5999;

		// hand-written asm to make sure we don't do something wasteful here
		__asm
		{
			push eax
			push ebx
			push edx
			mov ax,word ptr secs
			mov bx,60
			xor dx,dx
			div bx
			mov minutes,ax
			mov seconds,dx
			pop edx
			pop ebx
			pop eax
		};

		Video_SelectWritePlane(2);
		Font_Print2Dec(272, 102, minutes);
		Font_Print2Dec(296, 102, seconds);

		oldPlaybackSeconds = playbackSeconds;
	}
}

void UI_DrawSongTempo(void)
{
	Video_SelectWritePlane(2);

	if (module.useVblankTempo)
		Font_Print3Dec(32, 123, 125); // vblank = always 125 BPM ticks
	else
		Font_Print3Dec(32, 123, module.BPM);
}

void UI_DrawSongName(void)
{
	int32_t i, j, x;

	Video_SelectWritePlane(2);

	x = 104;
	for (i = 0; i < 20; i++, x += FONT_W)
	{
		char ch = module.name[i];
		if (ch == '\0')
		{
			for (j = i; j < 20; j++, x += FONT_W)
				Font_PrintChar(x, 102, '_');

			break;
		}

		Font_PrintChar(x, 102, ch);
	}
}

void UI_DrawSampleName(void)
{
	int32_t i, j, x;
	ptSmp_t *s = &samples[ui.currSample];

	Video_SelectWritePlane(2);

	x = 104;
	for (i = 0; i < 22; i++, x += FONT_W)
	{
		char ch = s->name[i];
		if (ch == '\0')
		{
			for (j = i; j < 22; j++, x += FONT_W)
				Font_PrintChar(x, 113, '_');

			break;
		}

		Font_PrintChar(x, 113, ch);
	}
}

void UI_DrawEditSkip(void)
{
	Video_SelectWritePlane(2);
	Font_PrintChar(24, 102, '0' + ui.editSkip);
}

static void SetEditSkip(uint8_t editskip)
{
	editSkipMsg[11] = '0' + editskip;
	UI_ShowTemporaryStatusText(editSkipMsg, 60);
	ui.editSkip = editskip;
	UI_DrawEditSkip();
}

// handles key presses for when askbox is not open
void UI_HandleKeyPress(uint8_t scancode)
{
	int8_t tmp8;

	switch (scancode)
	{
		case KEY_RSHIFT: Replayer_Play(PLAY_MODE_RECORD, 0);  break;
		case KEY_RALT:   Replayer_Play(PLAY_MODE_SONG, 0);    break;
		case KEY_RCTRL:  Replayer_Play(PLAY_MODE_PATTERN, 0); break;

		case KEY_ENTER:
		{
			Replayer_StepPlayForwards();
			ui.redrawPatternFlag = true;
		}
		break;

		case KEY_BACKSPACE:
		{
			Replayer_StepPlayBackwards();
			ui.redrawPatternFlag = true;
		}
		break;

		case KEY_F6:
		{
			if (keyb.shiftPressed)
			{
				ui.f6Pos = module.row;
				UI_ShowTemporaryStatusText("POSITION SET", 60);
			}
			else if (keyb.ctrlPressed)
			{
				Replayer_Play(PLAY_MODE_RECORD, ui.f6Pos);
			}
			else if (keyb.amigaPressed)
			{
				Replayer_Play(PLAY_MODE_SONG, ui.f6Pos);
			}
			else if (keyb.altPressed)
			{
				Replayer_Play(PLAY_MODE_PATTERN, ui.f6Pos);
			}
		}
		break;

		case KEY_F7:
		{
			if (keyb.shiftPressed)
			{
				ui.f7Pos = module.row;
				UI_ShowTemporaryStatusText("POSITION SET", 60);
			}
			else if (keyb.ctrlPressed)
			{
				Replayer_Play(PLAY_MODE_RECORD, ui.f7Pos);
			}
			else if (keyb.amigaPressed)
			{
				Replayer_Play(PLAY_MODE_SONG, ui.f7Pos);
			}
			else if (keyb.altPressed)
			{
				Replayer_Play(PLAY_MODE_PATTERN, ui.f7Pos);
			}
		}
		break;

		case KEY_F8:
		{
			if (keyb.shiftPressed)
			{
				ui.f8Pos = module.row;
				UI_ShowTemporaryStatusText("POSITION SET", 60);
			}
			else if (keyb.ctrlPressed)
			{
				Replayer_Play(PLAY_MODE_RECORD, ui.f8Pos);
			}
			else if (keyb.amigaPressed)
			{
				Replayer_Play(PLAY_MODE_SONG, ui.f8Pos);
			}
			else if (keyb.altPressed)
			{
				Replayer_Play(PLAY_MODE_PATTERN, ui.f8Pos);
			}
		}
		break;

		case KEY_F9:
		{
			if (keyb.shiftPressed)
			{
				ui.f9Pos = module.row;
				UI_ShowTemporaryStatusText("POSITION SET", 60);
			}
			else if (keyb.ctrlPressed)
			{
				Replayer_Play(PLAY_MODE_RECORD, ui.f9Pos);
			}
			else if (keyb.amigaPressed)
			{
				Replayer_Play(PLAY_MODE_SONG, ui.f9Pos);
			}
			else if (keyb.altPressed)
			{
				Replayer_Play(PLAY_MODE_PATTERN, ui.f9Pos);
			}
		}
		break;

		case KEY_F10:
		{
			if (keyb.shiftPressed)
			{
				ui.f10Pos = module.row;
				UI_ShowTemporaryStatusText("POSITION SET", 60);
			}
			else if (keyb.ctrlPressed)
			{
				Replayer_Play(PLAY_MODE_RECORD, ui.f10Pos);
			}
			else if (keyb.amigaPressed)
			{
				Replayer_Play(PLAY_MODE_SONG, ui.f10Pos);
			}
			else if (keyb.altPressed)
			{
				Replayer_Play(PLAY_MODE_PATTERN, ui.f10Pos);
			}
		}
		break;
		
		case KEY_F11:
		{
			if (keyb.shiftPressed)
			{
				if (!stereoAudio)
				{
					UI_ShowTemporaryStatusText("MONO ONLY!", 60);
				}
				else
				{
					char buf[32];
					
					stereoSeparation -= 10;
					if (stereoSeparation < 0)
						stereoSeparation = 0;

					Audio_UpdateStereoSeparation();

					sprintf(buf, "STEREO SEP.: %d%%\n", stereoSeparation);
					UI_ShowTemporaryStatusText(buf, 60);
				}
			}
			else
			{
				module.useVblankTempo ^= 1;
				_disable();
				Replayer_SetBPM(module.BPM);
				_enable();
				ui.redrawBPMFlag = true;
				ui.redrawTempoModeFlag = true;
			}
		}
		break;

		case KEY_F12:
		{
			if (keyb.shiftPressed)
			{
				if (!stereoAudio)
				{
					UI_ShowTemporaryStatusText("MONO ONLY!", 60);
				}
				else
				{
					char buf[32];
					
					stereoSeparation += 10;
					if (stereoSeparation > 100)
						stereoSeparation = 100;
						
					Audio_UpdateStereoSeparation();

					sprintf(buf, "STEREO SEP.: %d%%\n", stereoSeparation);
					UI_ShowTemporaryStatusText(buf, 60);
				}
			}
			else
			{
				audioAmiga500 ^= 1;
				Audio_ClearLpFilterState();
				
				if (audioAmiga500)
					UI_ShowTemporaryStatusText("AUDIO: A500", 60);
				else
					UI_ShowTemporaryStatusText("AUDIO: A1200", 60);
			}
		}
		break;

		case KEY_UP:
		{
			if (ui.diskOpShown)
			{
				DiskOp_ScrollUp();
			}
			else
			{
				if (!Replayer_SongPlaying)
				{
					module.row = (module.row - 1) & 63;
					ui.redrawPatternFlag = true;
				}
			}
		}
		break;

		case KEY_DOWN:
		{
			if (ui.diskOpShown)
			{
				DiskOp_ScrollDown();
			}
			else
			{
				if (!Replayer_SongPlaying)
				{
					module.row = (module.row + 1) & 63;
					ui.redrawPatternFlag = true;
				}
			}
		}
		break;

		case KEY_PAGEUP:
		{
			if (!Replayer_SongPlaying)
			{
				if (module.row >= 16)
				{
					module.row -= 16;
					ui.redrawPatternFlag = true;
				}
				else if (module.row != 0)
				{
					module.row = 0;
					ui.redrawPatternFlag = true;
				}
			}
		}
		break;
		
		case KEY_PAGEDOWN:
		{
			if (!Replayer_SongPlaying)
			{
				if (module.row <= 63-16)
				{
					module.row += 16;
					ui.redrawPatternFlag = true;
				}
				else if (module.row != 63)
				{
					module.row = 63;
					ui.redrawPatternFlag = true;
				}
			}
		}
		break;

		case KEY_HOME:
		{
			if (ui.diskOpShown)
			{
				DiskOp_ScrollToTop();
			}
			else
			{
				if (!Replayer_SongPlaying)
				{
					if (module.row != 0)
					{
						module.row = 0;
						ui.redrawPatternFlag = true;
					}
				}
			}
		}
		break;

		case KEY_END:
		{
			if (ui.diskOpShown)
			{
				DiskOp_ScrollToBottom();
			}
			else
			{
				if (!Replayer_SongPlaying)
				{
					if (module.row != 63)
					{
						module.row = 63;
						ui.redrawPatternFlag = true;
					}
				}
			}
		}
		break;

		case KEY_LEFT:
		{
			tmp8 = ui.patt.cursorPos - 1;
			if (tmp8 < 0)
				tmp8 = 23;
				
			Pattern_SetCursorPos(tmp8);
		}
		break;

		case KEY_RIGHT:
		{
			tmp8 = ui.patt.cursorPos + 1;
			if (tmp8 > 23)
				tmp8 = 0;

			Pattern_SetCursorPos(tmp8);
		}
		break;

		case KEY_TAB:
		{
			if (keyb.shiftPressed)
			{
				tmp8 = ((ui.patt.cursorPos + 5) / 6) - 1;
				if (tmp8 < 0)
					tmp8 = 3;
			}
			else
			{
				tmp8 = (ui.patt.cursorPos / 6) + 1;
				if (tmp8 == 4)
					tmp8 = 0;
			}

			Pattern_SetCursorPos(tmp8 * 6);
		}
		break;
	
		case KEY_D:
		{
			if (keyb.altPressed)
				DiskOp_Toggle();
		}
		break;
		
		case KEY_S:
		{
			if (keyb.altPressed)
				Sampler_Toggle();
		}
		break;
		
		case KEY_0:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(0);
		}
		break;

		case KEY_1:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(1);
		}
		break;

		case KEY_2:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(2);
		}
		break;

		case KEY_3:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(3);
		}
		break;

		case KEY_4:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(4);
		}
		break;

		case KEY_5:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(5);
		}
		break;

		case KEY_6:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(6);
		}
		break;

		case KEY_7:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(7);
		}
		break;

		case KEY_8:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(8);
		}
		break;

		case KEY_9:
		{
			if (keyb.ctrlPressed)
				SetEditSkip(9);
		}
		break;

		case KEY_ESC:
		{
			if (ui.diskOpShown)
			{
				DiskOp_Toggle();
			}
			else if (ui.samplerScreenShown)
			{
				Sampler_Toggle();
			}
			else
			{
				//if (AskBox("REALLY QUIT ?"))
					trackerRunning = false;
			}
		}
		break;

		case KEY_SPACE:
		{
			if (Replayer_SongPlaying)
			{
				Replayer_Stop();
			}
			else
			{
				if (Replayer_PlayMode != PLAY_MODE_EDIT)
					Replayer_PlayMode = PLAY_MODE_EDIT;
				else
					Replayer_PlayMode = PLAY_MODE_IDLE;
			}

			Mouse_UpdateColor();
		}
		break;
		
		case KEY_A:
		{
			if (keyb.ctrlPressed)
			{
				activeChannels ^= 1 << ((ui.patt.cursorPos / 6) & 3);		
				
				if (!ui.diskOpShown)
					GFX_DrawChannelMutes();
			}
		}
		break;
		
		case KEY_C:
		{
			if (keyb.altPressed)
			{
				activeChannels ^= 4;
				if (!ui.diskOpShown)
					GFX_DrawChannelMutes();
			}
		}
		break;

		case KEY_V:
		{
			if (keyb.altPressed)
			{
				activeChannels ^= 8;
				if (!ui.diskOpShown)
					GFX_DrawChannelMutes();
			}
		}
		break;	

		case KEY_Q:
		{
			if (keyb.ctrlPressed)
			{
				activeChannels = 15;
				if (!ui.diskOpShown)
					GFX_DrawChannelMutes();
			}
		}
		break;		

		case KEY_R:
		{
			if (keyb.ctrlPressed)
			{
				showRasterBars ^= 1;
				
				_disable();
				Video_RestoreBackgroundColor();
				Video_RestoreBackgroundColor();
				_enable();
			}
		}
		break;
		
		case KEY_X:
		{
			if (keyb.altPressed)
			{
				activeChannels ^= 2;
				if (!ui.diskOpShown)
					GFX_DrawChannelMutes();
			}
		}
		break;	

		case KEY_Z:
		{
			if (keyb.altPressed)
			{
				activeChannels ^= 1;
				if (!ui.diskOpShown)
					GFX_DrawChannelMutes();
			}
		}
		break;

		default: break;
	}
}

void UI_DrawTopScreen(void)
{
	GFX_DrawTopMainScreenBitmap();
	
	// some stuff has a zero-pad that never changes, draw them now
	Video_SelectWritePlane(2);
	Font_PrintChar(64,  3, '0');
	Font_PrintChar(64, 14, '0');
	Font_PrintChar(64, 25, '0');
	
	GFX_DrawChannelMutes();
	Scopes_ResetStates();
	UI_DrawSampleInfo();
	UI_DrawSongLength();
	UI_DrawPosInfo();
	UI_HandleUpdates();
}

void UI_DrawMiddleScreen(void)
{
	GFX_DrawMiddleMainScreenBitmap();
	UI_DrawEditSkip();
	UI_DrawSongName();
	UI_DrawSampleName();

	oldPlaybackSeconds = UINT32_MAX; // force update of playback timer
	UI_DrawPlaybackTimer();
}

void UI_DrawBottomScreen(void)
{
	if (ui.samplerScreenShown)
	{
		GFX_DrawSamplerScreenBitmap();
	}
	else
	{
		GFX_DrawPatternEditorBitmap();
		Pattern_RedrawCursor();
	}

	UI_DrawModSize();
	UI_DrawTempoMode();

	if (!ui.samplerScreenShown)
	{
		UI_DrawSongTempo();

		Pattern_ClearCache();
		ui.redrawPatternFlag = true;
	}

	UI_DrawStatusText();
	UI_HandleUpdates();
}

static void notImplemented(void)
{
	UI_Error("NOT IMPLEMENTED !");
}

static void songPosUp(void)
{
	int16_t tmp16;
	_disable();
	tmp16 = module.order;
	tmp16 += mouse.rightButtonDown ? 10 : 1;
	if (tmp16 > 127)
		tmp16 = 127;
	Replayer_SetPos((uint8_t)tmp16);
	_enable();
}

static void songPosDown(void)
{
	int16_t tmp16;
	_disable();
	tmp16 = module.order;
	tmp16 -= mouse.rightButtonDown ? 10 : 1;
	if (tmp16 < 0)
		tmp16 = 0;
	Replayer_SetPos((uint8_t)tmp16);
	_enable();
}

static void patternUp(void)
{
	int16_t tmp16;
	_disable();
	tmp16 = module.orderList[module.order];
	tmp16 += mouse.rightButtonDown ? 10 : 1;
	if (tmp16 > 99)
		tmp16 = 99;
	module.pattern = module.orderList[module.order] = (uint8_t)tmp16;
	_enable();

	Video_SelectWritePlane(2);
	Font_Print3Dec(72, 14, module.pattern);
}

static void patternDown(void)
{
	int16_t tmp16;
	_disable();
	tmp16 = module.orderList[module.order];
	tmp16 -= mouse.rightButtonDown ? 10 : 1;
	if (tmp16 < 0)
		tmp16 = 0;
	module.pattern = module.orderList[module.order] = (uint8_t)tmp16;
	_enable();

	Video_SelectWritePlane(2);
	Font_Print3Dec(72, 14, module.pattern);
}

static void songLenUp(void)
{
	int16_t tmp16;
	_disable();
	tmp16 = module.length;
	tmp16 += mouse.rightButtonDown ? 10 : 1;
	if (tmp16 > 128)
		tmp16 = 128;
	module.length = (uint8_t)tmp16;
	_enable();
	UI_DrawSongLength();
}

static void songLenDown(void)
{
	int16_t tmp16;
	_disable();
	tmp16 = module.length;
	tmp16 -= mouse.rightButtonDown ? 10 : 1;
	if (tmp16 < 1)
		tmp16 = 1;
	module.length = (uint8_t)tmp16;
	_enable();
	UI_DrawSongLength();
}

static void finetuneUp(void)
{
	uint8_t tmp8;
	ptSmp_t *s = &samples[ui.currSample];
	
	if (mouse.rightButtonDown)
	{
		s->finetune = 0;
	}
	else
	{
		_disable();
		tmp8 = s->finetune & 0xF;
		if (tmp8 != 7)
			s->finetune = (tmp8 + 1) & 0xF;
		_enable();
	}

	Video_SelectWritePlane(2);
	Font_Print(80, 36, ftuneStrTab[s->finetune & 0xF]);
}

static void finetuneDown(void)
{
	uint8_t tmp8;
	ptSmp_t *s = &samples[ui.currSample];

	if (mouse.rightButtonDown)
	{
		s->finetune = 0;
	}
	else
	{
		_disable();
		tmp8 = s->finetune & 0xF;
		if (tmp8 != 8)
			s->finetune = (tmp8 - 1) & 0xF;
		_enable();
	}

	Video_SelectWritePlane(2);
	Font_Print(80, 36, ftuneStrTab[s->finetune & 0xF]);
}

static void sampleUp(void)
{
	// TODO: mouse.rightButtonDown = set sample-zero
	
	if (ui.currSample < 30)
	{
		ui.currSample++;
		Sampler_ShowSampleInfo();
		if (ui.samplerScreenShown)
			Sampler_DisplaySample();
	}
}

static void sampleDown(void)
{
	// TODO: mouse.rightButtonDown = set sample-zero

	if (ui.currSample > 0)
	{
		ui.currSample--;
		Sampler_ShowSampleInfo();
		if (ui.samplerScreenShown)
			Sampler_DisplaySample();
	}
}

static void sampleVolumeUp(void)
{
	ptSmp_t *s = &samples[ui.currSample];

	_disable();
	s->volume += mouse.rightButtonDown ? 16 : 1;
	if (s->volume > 64)
		s->volume = 64;
	_enable();
	
	Video_SelectWritePlane(2);
	Font_PrintHexWord(64, 58, s->volume);
}

static void sampleVolumeDown(void)
{
	ptSmp_t *s = &samples[ui.currSample];

	_disable();
	s->volume -= mouse.rightButtonDown ? 16 : 1;
	if ((int8_t)s->volume < 0)
		s->volume = 0;
	_enable();
	
	Video_SelectWritePlane(2);
	Font_PrintHexWord(64, 58, s->volume);
}

static void repeatUp(void)
{
	ptSmp_t *s = &samples[ui.currSample];

	if (s->length == 0)
	{
		s->loopStart = 0;
		return;
	}

	_disable();
	s->loopStart += mouse.rightButtonDown ? 8 : 1;
	if (s->loopStart+s->loopLength > s->length)
		s->loopStart = s->length - s->loopLength;
	_enable();
	
	Video_SelectWritePlane(2);
	Font_PrintHexWord(64, 80, s->loopStart << 1);
	
	Replayer_UpdateRepeats(s);
	if (ui.samplerScreenShown)	
		Sampler_RedrawWaveform();
}

static void repeatDown(void)
{
	ptSmp_t *s = &samples[ui.currSample];

	_disable();
	s->loopStart -= mouse.rightButtonDown ? 8 : 1;
	if ((int32_t)(int16_t)s->loopStart < 0)
		s->loopStart = 0;
	_enable();
	
	Video_SelectWritePlane(2);
	Font_PrintHexWord(64, 80, s->loopStart << 1);
	
	Replayer_UpdateRepeats(s);
	if (ui.samplerScreenShown)
		Sampler_RedrawWaveform();
}

static void replenUp(void)
{
	ptSmp_t *s = &samples[ui.currSample];

	if (s->length == 0)
	{
		s->loopLength = 1;
		return;
	}

	_disable();
	s->loopLength += mouse.rightButtonDown ? 8 : 1;
	if (s->loopStart+s->loopLength > s->length)
		s->loopLength = s->length - s->loopStart;
		
	if ((int32_t)(int16_t)s->loopLength < 1)
		s->loopLength = 1;
	_enable();
	
	Video_SelectWritePlane(2);
	Font_PrintHexWord(64, 91, s->loopLength << 1);
	
	Replayer_UpdateRepeats(s);
	if (ui.samplerScreenShown)	
		Sampler_RedrawWaveform();
}

static void replenDown(void)
{
	ptSmp_t *s = &samples[ui.currSample];

	_disable();
	s->loopLength -= mouse.rightButtonDown ? 8 : 1;
	if ((int32_t)(int16_t)s->loopLength < 1)
		s->loopLength = 1;
	_enable();
	
	Video_SelectWritePlane(2);
	Font_PrintHexWord(64, 91, s->loopLength << 1);
	
	Replayer_UpdateRepeats(s);
	if (ui.samplerScreenShown)
		Sampler_RedrawWaveform();
}

static void increaseBPM(void)
{
	volatile int16_t bpm = module.BPM;

	bpm += mouse.rightButtonDown ? 10 : 1;
	if (bpm > PT_MAX_BPM)
		bpm = PT_MAX_BPM;
		
	module.BPM = (uint8_t)bpm;
	Replayer_SetBPM(module.BPM);
	UI_DrawSongTempo();
}

static void decreaseBPM(void)
{
	volatile int16_t bpm = module.BPM;

	bpm -= mouse.rightButtonDown ? 10 : 1;
	if (bpm < PT_MIN_BPM)
		bpm = PT_MIN_BPM;
		
	module.BPM = (uint8_t)bpm;
	Replayer_SetBPM(module.BPM);
	UI_DrawSongTempo();
}

// handles mouse button presses when askbox is not open
void UI_HandleMouseClick(void)
{
	int8_t tmp8;
	ptSmp_t *s = &samples[ui.currSample];

	if (mouse.x == 0 && mouse.y == 0)
	{
		if (AskBox("REALLY QUIT ?"))
			trackerRunning = false;
			
		return;
	}

	// -------------------------------------------------
	//                   TOP SCREEN
	// -------------------------------------------------
	if (ui.diskOpShown)
	{
		if (DiskOp_HandleListSelection())
			return;

		if (mouse.y < 11)
		{
			if (mouse.x < 80)
			{
				// PARENT
				DiskOp_SetPath("..");
				return;
			}
			else if (mouse.x < 146)
			{
				// PACK
				notImplemented();
				return;
			}
			else if (mouse.x < 238)
			{
				// LOAD MODULE (mode)
				notImplemented();
				return;
			}
			else
			{
				// SAVE MODULE
				notImplemented();
				return;
			}
		}
		else if (mouse.y < 22)
		{
			if (mouse.x < 80)
			{
				// REFRESH
				DiskOp_Refresh();
				return;
			}
			else if (mouse.x < 146)
			{
				// SAVE
				notImplemented();
				return;
			}
			else if (mouse.x < 238)
			{
				// LOAD SAMPLE (mode)
				notImplemented();
				return;
			}
			else
			{
				// SAVE SAMPLE
				notImplemented();
				return;
			}
		}
		else if (mouse.y < 33)
		{
			if (mouse.x < 308)
			{
				// Set disk op. path
				notImplemented();
				return;
			}
			else
			{
				// Scroll to top
				DiskOp_ScrollToTop();
				return;
			}
		}
		else
		{
			if (mouse.x >= 308)
			{
				if (mouse.y < 40)
				{
					DiskOp_ScrollUp();
					Mouse_SetupGUIButtonRepeat(308, 320, 33, 40, DiskOp_ScrollUp, 16, 1);
					return;
				}
				else if (mouse.y < 81)
				{
					// Exit
					DiskOp_Exit();
					return;
				}
				else if (mouse.y < 90)
				{
					DiskOp_ScrollDown();
					Mouse_SetupGUIButtonRepeat(308, 320, 81, 90, DiskOp_ScrollDown, 16, 1);
					return;
				}
				else if (mouse.y < 99)
				{
					// Scroll to bottom
					DiskOp_ScrollToBottom();
					return;
				}
			}
		}
	}
	else 
	{
		if (mouse.x >= 120 && mouse.y >= 44 && mouse.y < 99)
			return; // scopes area = do nothing in PT4DOS (no spectrum analyzer)
		
		if (mouse.y < 11)
		{
			if (mouse.x < 40)
			{
				// POS
				notImplemented();
				return;
			}
			else if (mouse.x < 51)
			{
				// I (insert pattern)
				notImplemented();
				return;
			}
			else if (mouse.x < 62)
			{
				// D (delete pattern)
				notImplemented();
				return;
			}
			else if (mouse.x < 98)
			{
				// Set current position
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				songPosUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 0, 11, songPosUp, 16, 6);
				return;
			}
			else if (mouse.x < 120)
			{
				songPosDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 0, 11, songPosDown, 16, 6);
				return;
			}
			else if (mouse.x < 182)
			{
				// PLAY
				if (mouse.rightButtonDown)
					Replayer_Play(PLAY_MODE_SONG, module.row);
				else
					Replayer_Play(PLAY_MODE_SONG, 0);

				return;
			}
			else if (mouse.x < 244)
			{
				// STOP
				Replayer_Stop();
				Mouse_UpdateColor();
				return;
			}
			else if (mouse.x < 306)
			{
				// PLST
				notImplemented();
				return;
			}
			else
			{
				// Mute channel #1
				if (mouse.rightButtonDown) // solo
					activeChannels = 1;
				else
					activeChannels ^= 1;
				GFX_DrawChannelMutes();
				return;
			}
		}
		else if (mouse.y < 22)
		{
			if (mouse.x < 62)
				return;

			if (mouse.x < 98)
			{
				// Set current pattern
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				patternUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 11, 22, patternUp, 16, 6);
				return;
			}
			else if (mouse.x < 120)
			{
				patternDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 11, 22, patternDown, 16, 6);
				return;
			}
			else if (mouse.x < 182)
			{
				// PATTERN
				if (mouse.rightButtonDown)
					Replayer_Play(PLAY_MODE_PATTERN, module.row);
				else
					Replayer_Play(PLAY_MODE_PATTERN, 0);

				return;
			}
			else if (mouse.x < 244)
			{
				// CLEAR
				notImplemented();
				return;
			}
			else if (mouse.x < 306)
			{
				// PSET-ED
				notImplemented();
				return;
			}
			else
			{
				// Mute channel #2
				if (mouse.rightButtonDown) // solo
					activeChannels = 2;
				else
					activeChannels ^= 2;
				GFX_DrawChannelMutes();
				return;
			}
		}
		else if (mouse.y < 33)
		{
			if (mouse.x < 62)
				return;

			if (mouse.x < 98)
			{
				// Set song length
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				songLenUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 22, 33, songLenUp, 16, 6);
				return;
			}
			else if (mouse.x < 120)
			{
				songLenDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 22, 33, songLenDown, 16, 6);
				return;
			}
			else if (mouse.x < 182)
			{
				// EDIT

				if (Replayer_SongPlaying)
					Replayer_Stop();
				
				Replayer_PlayMode = PLAY_MODE_EDIT;
				Mouse_UpdateColor();

				return;
			}
			else if (mouse.x < 244)
			{
				// EDIT OP.
				notImplemented();
				return;
			}
			else if (mouse.x < 306)
			{
				// SETUP
				notImplemented();
				return;
			}
			else
			{
				// Mute channel #3
				if (mouse.rightButtonDown) // solo
					activeChannels = 4;
				else
					activeChannels ^= 4;
				GFX_DrawChannelMutes();
				return;
			}
		}
		else if (mouse.y < 44)
		{
			if (mouse.x < 98)
				return;

			if (mouse.x < 109)
			{
				finetuneUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 33, 44, finetuneUp, 16, 4);
				return;
			}
			else if (mouse.x < 120)
			{
				finetuneDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 33, 44, finetuneDown, 16, 4);
				return;
			}
			else if (mouse.x < 182)
			{
				// RECORD
				if (mouse.rightButtonDown)
					Replayer_Play(PLAY_MODE_RECORD, module.row);
				else
					Replayer_Play(PLAY_MODE_RECORD, 0);

				return;
			}
			else if (mouse.x < 244)
			{
				// DISK OP.
				DiskOp_Open();
				return;
			}
			else if (mouse.x < 306)
			{
				// SAMPLER
				Sampler_Toggle();
				return;
			}
			else
			{
				// Mute channel #4	
				if (mouse.rightButtonDown) // solo
					activeChannels = 8;
				else
					activeChannels ^= 8;
				GFX_DrawChannelMutes();
				return;
			}
		}
		else if (mouse.y < 55)
		{
			if (mouse.x < 62 || mouse.x >= 120)
				return;
				
			if (mouse.x < 98)
			{
				// Set current sample
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				sampleUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 44, 55, sampleUp, 16, 4);
				return;
			}
			else
			{
				sampleDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 44, 55, sampleDown, 16, 4);
				return;
			}
		}
		else if (mouse.y < 66)
		{
			if (mouse.x < 62 || mouse.x >= 120)
				return;

			if (mouse.x < 98)
			{
				// Set sample volume
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				sampleVolumeUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 55, 66, sampleVolumeUp, 16, 4);
				return;
			}
			else
			{
				sampleVolumeDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 55, 66, sampleVolumeDown, 16, 4);
				return;
			}
		}
		else if (mouse.y < 77)
		{
			if (mouse.x < 62 || mouse.x >= 120)
				return;

			if (mouse.x < 98)
			{
				// Set sample length
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				// Sample length up
				notImplemented();
				return;
			}
			else
			{
				// Sample length down
				notImplemented();
				return;
			}
		}
		else if (mouse.y < 88)
		{
			if (mouse.x < 62 || mouse.x >= 120)
				return;

			if (mouse.x < 98)
			{
				// Set sample repeat
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				repeatUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 77, 88, repeatUp, 16, 0);
				return;
			}
			else
			{
				repeatDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 77, 88, repeatDown, 16, 0);
				return;
			}
		}
		else if (mouse.y < 99)
		{
			if (mouse.x < 62 || mouse.x >= 120)
				return;

			if (mouse.x < 98)
			{
				// Set sample replen
				notImplemented();
				return;
			}
			else if (mouse.x < 109)
			{
				replenUp();
				Mouse_SetupGUIButtonRepeat(98, 109, 88, 99, replenUp, 16, 0);	
				return;
			}
			else
			{
				replenDown();
				Mouse_SetupGUIButtonRepeat(109, 120, 88, 99, replenDown, 16, 0);
				return;
			}
		}
	}

	// -------------------------------------------------
	//   MIDDLE SCREEN (song/sample name, LOAD button)
	// -------------------------------------------------
	// TODO


	// -------------------------------------------------
	//                  BOTTOM SCREEN
	// -------------------------------------------------
	if (ui.samplerScreenShown)
	{
		// TODO: zero-sample test
		ptSmp_t *s = &samples[ui.currSample];
		
		if (mouse.y >= 138 && mouse.y < 186)
		{
			if (mouse.y < 142) // check loop pins
			{
				if (sampler.loopStartPos >= 0) // left loop pin
				{
					int32_t oldLoopEnd = s->loopStart + s->loopLength;
					if (mouse.x >= sampler.loopStartPos && mouse.x < sampler.loopStartPos+4)
					{
						int32_t oldMouseX = mouse.x;
						while (mouse.leftButtonDown)
						{
							Video_WaitForVblank();
							Sampler_EraseOldSamplingLine();
							Keyb_HandleInput();
							UI_AmigaVblankHandler();
							UI_HandleUpdates();
					
							if (mouse.x != oldMouseX)
							{
								oldMouseX = mouse.x;
								Sampler_SetLoopStartFromX(s, mouse.x, oldLoopEnd);
							}
							
							Sampler_WriteSamplingLine();
						}
	
						return;
					}
				}
				
				if (sampler.loopEndPos >= 0) // right loop pin
				{
					if (mouse.x >= sampler.loopEndPos && mouse.x < sampler.loopEndPos+4)
					{
						int32_t oldMouseX = mouse.x;
						while (mouse.leftButtonDown)
						{
							Video_WaitForVblank();
							Sampler_EraseOldSamplingLine();
							Keyb_HandleInput();
							UI_AmigaVblankHandler();
							UI_HandleUpdates();
							
							if (mouse.x != oldMouseX)
							{
								oldMouseX = mouse.x;
								Sampler_SetLoopEndFromX(s, mouse.x);
							}
							
							Sampler_WriteSamplingLine();
						}

						return;
					}
				}
			}
			
			if (mouse.x >= 3 && mouse.x < 317) // sample data marking
			{
				int32_t oldMarkStart, mx, lastSamPos = mouse.x;

				sampler.markStart = sampler.markEnd = lastSamPos;
				Sampler_RedrawWaveform();
				Sampler_MarkToOffset();
				sampler.samplePos = sampler.markEndOfs;
				//EditOp_ShowPos();

				while (mouse.leftButtonDown)
				{
					Video_WaitForVblank();
					Sampler_EraseOldSamplingLine();
					Keyb_HandleInput();
					UI_AmigaVblankHandler();
					UI_HandleUpdates();
					
					mx = CLAMP(mouse.x, 3, 316);
					if (mx != lastSamPos)
					{
						lastSamPos = sampler.markEnd = mx;
						Sampler_RedrawWaveform();
						Sampler_MarkToOffset();
						sampler.samplePos = sampler.markEndOfs;
						//EditOp_ShowPos();
					}
					
					Sampler_WriteSamplingLine();
				}
				
				if (sampler.markStart > sampler.markEnd) // swap mark start/end if needed
				{
					oldMarkStart = sampler.markStart;
					sampler.markStart = sampler.markEnd;
					sampler.markEnd = oldMarkStart;
					Sampler_MarkToOffset();
				}
			}

			return;
		}
		else if (mouse.y >= 124 && mouse.y < 135)
		{
			if (mouse.x >= 6 && mouse.x < 26)
			{
				// EXIT
				Sampler_Toggle();
				return;
			}
		}
		// TODO
	}
	else
	{
		// pattern editor
		
		if (mouse.y >= 121 && mouse.y < 138)
		{
			if (mouse.x < 26)
			{
				// Select edit pattern
				notImplemented();
				return;
			}
			else if (mouse.x < 44)
			{
				increaseBPM();
				Mouse_SetupGUIButtonRepeat(26, 44, 121, 138, increaseBPM, 16, 2);
				return;
			}
			else if (mouse.x < 60)
			{
				decreaseBPM();
				Mouse_SetupGUIButtonRepeat(44, 60, 121, 138, decreaseBPM, 16, 2);
				return;
			}
		}

		// TODO (pattern editor, select pattern)
	}
}
